"""Functions to load sample data."""
import geopandas as gpd
import pooch

REGISTRY = pooch.create(
    path=pooch.os_cache("pandamesh"),
    base_url="https://raw.githubusercontent.com/Deltares/pandamesh/main/data/",
    version=None,
    version_dev="main",
    env="PANDAMESH_DATA_DIR",
    registry={
        "provinces-nl.geojson": "7539318974d1d78f35e4c2987287aa81f5ff505f444a2e0f340d804f57c0f8e3",
        "south-america.geojson": "337746351d15a83d5d41f1cecd30aa40b1698eb7587a4e412c511af89f82e49c",
    },
)


def provinces_nl():
    """Return the provinces (including water bodies) of the Netherlands as a GeoDataframe."""
    fname = REGISTRY.fetch("provinces-nl.geojson")
    return gpd.read_file(fname)


def south_america():
    fname = REGISTRY.fetch("south-america.geojson")
    return gpd.read_file(fname)



import re
import struct
from pathlib import Path
from typing import NamedTuple

import geopandas as gpd
import numpy as np
import pytest
import shapely.geometry as sg
import xarray as xr

from pandamesh import gmsh_fields as gf
from pandamesh.gmsh_mesher import GmshMesher


@pytest.fixture(scope="function")
def gdf():
    polygon = sg.Polygon(
        [
            [0.0, 0.0],
            [10.0, 0.0],
            [10.0, 10.0],
            [0.0, 10.0],
        ]
    )
    return gpd.GeoDataFrame(geometry=[polygon], data={"cellsize": [10.0]})


def assert_refinement(mesh: gpd.GeoDataFrame):
    assert (mesh.area.max() / mesh.area.min()) > 5


class StructuredFieldContent(NamedTuple):
    cellsize: np.ndarray
    xmin: float
    ymin: float
    dx: float
    dy: float


def read_structured_field_file(
    path: Path
) -> tuple[np.ndarray, float, float, float, float]:
    with path.open("rb") as f:
        xmin, ymin, _ = struct.unpack("3d", f.read(24))  # 3 doubles, 8 bytes each
        dx, dy, _ = struct.unpack("3d", f.read(24))  # 3 doubles, 8 bytes each
        nrow, ncol, _ = struct.unpack("3i", f.read(12))  # 3 integers, 4 bytes each
        cellsize_data = f.read()  # Read the rest of the file

    cellsize = np.frombuffer(cellsize_data, dtype=np.float64).reshape((ncol, nrow))
    return StructuredFieldContent(cellsize, xmin, ymin, dx, dy)


def test_write_structured_field_file(tmp_path):
    cellsize = np.arange(6.0).reshape(2, 3)
    xmin = 5.0
    ymin = 15.0
    path = tmp_path / "a.dat"
    with pytest.raises(ValueError, match=re.escape("`cellsize` must be 2D")):
        gf.write_structured_field_file(path, cellsize[0], xmin, ymin, 1.0, 1.0)

    gf.write_structured_field_file(path, cellsize, xmin, ymin, 1.0, 1.0)
    assert path.exists()
    back = read_structured_field_file(path)
    assert back.dx == 1.0
    assert back.dy == 1.0
    assert back.xmin == 5.0
    assert back.ymin == 15.0
    # Gmsh expects it in column major order
    assert np.array_equal(back.cellsize, cellsize.T)

    path = tmp_path / "b.dat"
    gf.write_structured_field_file(path, cellsize, xmin, ymin, 1.0, -1.0)
    back = read_structured_field_file(path)
    assert back.dy == 1.0
    assert np.array_equal(back.cellsize[0, :], [3, 0])
    assert np.array_equal(back.cellsize[1, :], [4, 1])
    assert np.array_equal(back.cellsize[2, :], [5, 2])

    path = tmp_path / "c.dat"
    gf.write_structured_field_file(path, cellsize, xmin, ymin, -1.0, 1.0)
    back = read_structured_field_file(path)
    assert back.dx == 1.0
    assert np.array_equal(back.cellsize[0, :], [2, 5])
    assert np.array_equal(back.cellsize[1, :], [1, 4])
    assert np.array_equal(back.cellsize[2, :], [0, 3])


def test_math_eval_field(gdf):
    mesher = GmshMesher._force_init(gdf)

    with pytest.raises(TypeError):
        mesher.add_matheval_distance_field(1)

    field = gpd.GeoDataFrame(
        geometry=[sg.Point([5.0, 5.0])],
    )
    # missing columns
    with pytest.raises(ValueError):
        mesher.add_matheval_distance_field(field)

    field["spacing"] = np.nan
    field["function"] = "dist + 0.1"

    with pytest.raises(ValueError, match="distance not in MathEval field function"):
        mesher.add_matheval_distance_field(field)

    field["function"] = "max(distance, 0.5)"
    mesher.add_matheval_distance_field(field)
    mesh = mesher.generate_geodataframe()

    field = mesher.fields[0]
    assert isinstance(field, gf.MathEvalField)

    # Test whether anything has been refined.
    assert_refinement(mesh)

    # Assert combination field has been generated
    assert mesher._combination_field is not None
    assert isinstance(mesher._combination_field, gf.CombinationField)

    mesher.clear_fields()
    assert len(mesher.fields) == 0
    assert mesher._combination_field is None


def test_threshold_field(gdf):
    mesher = GmshMesher._force_init(gdf)

    with pytest.raises(TypeError):
        mesher.add_threshold_distance_field(1)

    field = gpd.GeoDataFrame(
        geometry=[sg.Point([5.0, 5.0])],
    )
    # missing columns
    with pytest.raises(ValueError):
        mesher.add_threshold_distance_field(field)

    field["dist_min"] = 1.0
    field["dist_max"] = 2.5
    field["size_min"] = 0.2
    field["size_max"] = 2.5
    field["spacing"] = np.nan

    mesher.add_threshold_distance_field(field)
    mesh = mesher.generate_geodataframe()

    # Test whether anything has been refined.
    assert_refinement(mesh)

    # Should work WITH optional columns as well
    mesher.clear_fields()
    field["sigmoid"] = True
    field["stop_at_dist_max"] = True

    mesher.add_threshold_distance_field(field)

    field = mesher.fields[0]
    assert isinstance(field, gf.ThresholdField)

    mesh = mesher.generate_geodataframe()
    assert isinstance(mesh, gpd.GeoDataFrame)


def test_line_field(gdf):
    line = sg.LineString(
        [
            [3.0, -3.0],
            [3.0, 13.0],
        ]
    )
    field = gpd.GeoDataFrame(geometry=[line])
    field["dist_min"] = 2.0
    field["dist_max"] = 4.0
    field["size_min"] = 0.5
    field["size_max"] = 2.5
    field["spacing"] = 2.0

    mesher = GmshMesher._force_init(gdf)
    mesher.add_threshold_distance_field(field)

    field = mesher.fields[0]
    assert isinstance(field, gf.ThresholdField)

    mesh = mesher.generate_geodataframe()
    assert_refinement(mesh)


def test_polygon_field(gdf):
    square = sg.Polygon(
        [
            [3.0, 3.0],
            [7.0, 3.0],
            [7.0, 7.0],
            [3.0, 7.0],
        ]
    )
    field = gpd.GeoDataFrame(geometry=[square])
    field["dist_min"] = 0.5
    field["dist_max"] = 1.5
    field["size_min"] = 0.3
    field["size_max"] = 2.5
    field["spacing"] = 1.0

    mesher = GmshMesher._force_init(gdf)
    mesher.add_threshold_distance_field(field)

    field = mesher.fields[0]
    assert isinstance(field, gf.ThresholdField)

    mesh = mesher.generate_geodataframe()
    assert_refinement(mesh)


def test_add_structured_field(gdf):
    mesher = GmshMesher._force_init(gdf)

    y, x = np.meshgrid([1.0, 5.0, 9.0], [1.0, 5.0, 9.0], indexing="ij")
    distance_from_origin = np.sqrt((x * x + y * y))
    cellsize = np.log(distance_from_origin / distance_from_origin.min()) + 0.5

    with pytest.raises(ValueError, match=r"Minimum cellsize must be > 0, received:"):
        mesher.add_structured_field(
            cellsize=cellsize * -1,
            xmin=x.min(),
            ymin=y.min(),
            dx=1.0,
            dy=1.0,
        )

    mesher.add_structured_field(
        cellsize=cellsize,
        xmin=x.min(),
        ymin=y.min(),
        dx=1.0,
        dy=1.0,
    )

    field = mesher.fields[0]
    assert isinstance(field, gf.StructuredField)
    assert not field.set_outside_value
    assert field.outside_value == -1.0
    assert isinstance(field.path, Path)
    assert field.path.exists()

    mesh = mesher.generate_geodataframe()
    assert_refinement(mesh)

    mesher.clear_fields()
    mesher.add_structured_field(
        cellsize=cellsize,
        xmin=x.min(),
        ymin=y.min(),
        dx=1.0,
        dy=1.0,
        outside_value=100.0,
    )
    field = mesher.fields[0]
    assert isinstance(field, gf.StructuredField)
    assert field.set_outside_value
    assert field.outside_value == 100.0
    assert isinstance(field.path, Path)
    assert field.path.exists()


def test_add_structured_field_from_dataarray(gdf):
    mesher = GmshMesher._force_init(gdf)

    with pytest.raises(
        TypeError, match="da must be xr.DataArray, received instead: int"
    ):
        mesher.add_structured_field_from_dataarray(1)

    x = np.arange(1.0, 10.0)
    y = np.arange(1.0, 10.0)
    da = xr.DataArray(
        np.ones((y.size, x.size)), coords={"y": y, "x": x}, dims=("y", "x")
    )

    with pytest.raises(ValueError, match=re.escape('Dimensions must be ("y", "x")')):
        mesher.add_structured_field_from_dataarray(da.transpose())

    with pytest.raises(ValueError, match=r"Minimum cellsize must be > 0, received:"):
        mesher.add_structured_field_from_dataarray(da * -1)

    da_x = da.copy()
    da_x["x"] = da_x["x"] ** 2

    with pytest.raises(ValueError, match="da is not equidistant along x"):
        mesher.add_structured_field_from_dataarray(da_x)

    da_y = da.copy()
    da_y["y"] = da_y["y"] ** 2

    with pytest.raises(ValueError, match="da is not equidistant along y"):
        mesher.add_structured_field_from_dataarray(da_y)



import numpy as np

from pandamesh.common import FloatArray, IntArray


def plot(
    vertices: FloatArray,
    faces: IntArray,
    fill_value: int = -1,
    ax=None,
    facecolors: str = "lightgray",
    edgecolors: str = "black",
    **kwargs,
) -> None:
    """Plot an unstructured mesh"""
    import matplotlib.pyplot as plt
    from matplotlib.collections import PolyCollection

    if ax is None:
        _, ax = plt.subplots()

    nodes = vertices[faces]
    nodes[faces == fill_value] = np.nan
    collection = PolyCollection(
        nodes, facecolors=facecolors, edgecolors=edgecolors, **kwargs
    )
    ax.add_collection(collection)
    ax.set_aspect(1.0)
    ax.autoscale()



from __future__ import annotations

import functools
import operator
import warnings
from typing import Any, Sequence, Set, Tuple

import geopandas as gpd
import numpy as np
import pandas as pd
import shapely
from pandas.api.types import is_integer_dtype


class MaybeGmsh:  # pragma: no cover
    """Gmsh is an optional dependency."""

    def __init__(self):
        try:
            import gmsh

            self.gmsh = gmsh
            self.ok = True
            self.error = None
        except ImportError:
            self.gmsh = None
            self.ok = False
            self.error = ImportError("Gmsh is required for this functionality")
        except ValueError:
            self.gmsh = None
            self.ok = False
            self.error = RuntimeError(
                "Gmsh must run in the main thread of the interpreter"
            )

    def __getattr__(self, name: str):
        if self.ok:
            return getattr(self.gmsh, name)
        else:
            raise self.error


gmsh = MaybeGmsh()
BoolArray = np.ndarray
IntArray = np.ndarray
FloatArray = np.ndarray
GeometryArray = np.ndarray
coord_dtype = np.dtype([("x", np.float64), ("y", np.float64)])

POINT = shapely.GeometryType.POINT
LINESTRING = shapely.GeometryType.LINESTRING
LINEARRING = shapely.GeometryType.LINEARRING
POLYGON = shapely.GeometryType.POLYGON
GEOM_NAMES = {v: k for k, v in shapely.GeometryType.__members__.items()}


def repr(obj: Any) -> str:
    strings = [type(obj).__name__]
    for k, v in obj.__dict__.items():
        if k.startswith("_"):
            k = k[1:]
        if isinstance(v, np.ndarray):
            s = f"    {k} = np.ndarray with shape {v.shape}"
        else:
            s = f"    {k} = {v}"
        strings.append(s)
    return "\n".join(strings)


def flatten(seq: Sequence[Any]):
    return functools.reduce(operator.concat, seq)


def flatten_geometry(geom):
    """Recursively flatten geometry collections."""
    if hasattr(geom, "geoms"):
        return [g for subgeom in geom.geoms for g in flatten_geometry(subgeom)]
    else:
        return [geom]


def flatten_geometries(geometries: Sequence) -> GeometryArray:
    """Flatten geometry collections."""
    flattened = []
    for geom in geometries:
        flattened.extend(flatten_geometry(geom))
    return np.array(flattened)


def check_geodataframe(
    features: gpd.GeoDataFrame, required_columns: Set[str], check_index: bool = False
) -> None:
    if not isinstance(features, gpd.GeoDataFrame):
        raise TypeError(
            f"Expected GeoDataFrame, received instead: {type(features).__name__}"
        )

    if len(features) == 0:
        raise ValueError("Dataframe is empty")

    missing = required_columns - set(features.columns)
    if missing:
        raise ValueError(
            f"These column(s) are required but are missing: {', '.join(missing)}"
        )
    if check_index:
        if not is_integer_dtype(features.index):
            raise ValueError(
                f"geodataframe index is not integer typed, received: {features.index.dtype}"
            )
        if features.index.duplicated().any():
            raise ValueError(
                "geodataframe index contains duplicates, call .reset_index()"
            )


def intersecting_features(features, feature_type) -> Tuple[IntArray, IntArray]:
    tree = shapely.STRtree(geoms=features)
    if feature_type == "polygon":  # TODO: might not be necessary
        target = features.buffer(-1.0e-6)
    else:
        target = features
    i, j = tree.query(geometry=target, predicate="intersects")
    # Intersection matrix is symmetric, and contains i==j (diagonal)
    keep = j > i
    return i[keep], j[keep]


def check_intersection(features: gpd.GeoSeries, feature_type: str) -> None:
    if len(features) <= 1:
        return
    index_a, index_b = intersecting_features(features, feature_type)
    n_overlap = len(index_a)
    if n_overlap > 0:
        message = "\n".join([f"{a} with {b}" for a, b in zip(index_a, index_b)])
        raise ValueError(
            f"{n_overlap} cases of intersecting {feature_type} detected:\n{message}"
        )


def check_features(features: gpd.GeoSeries, feature_type) -> None:
    """Check whether features are simple."""
    are_simple = features.is_simple
    n_complex = (~are_simple).sum()
    if n_complex > 0:
        raise ValueError(
            f"{n_complex} cases of complex {feature_type} detected: these "
            " features contain self intersections"
        )
    return


def check_polygons(polygons: gpd.GeoSeries) -> None:
    check_features(polygons, "polygon")
    check_intersection(polygons, "polygon")
    return


def check_lines(lines: gpd.GeoSeries) -> None:
    check_features(lines, "lines")
    return


def compute_intersections(segment_linestrings, i: IntArray, j: IntArray):
    check = j > i
    i = i[check]
    j = j[check]

    intersections = shapely.intersection(
        segment_linestrings[i],
        segment_linestrings[j],
    )
    coordinates, index = shapely.get_coordinates(intersections, return_index=True)
    coordinates, unique_index = np.unique(coordinates, return_index=True, axis=0)
    index = index[unique_index]
    return i[index], j[index], coordinates


def linework_intersections(
    polygons: gpd.GeoSeries,
    lines: gpd.GeoSeries,
    tolerance: float,
) -> FloatArray:
    """Compute intersections between polygon boundary and line segments."""
    outer_rings = polygons.exterior
    inner_rings = np.array(flatten(polygons.interiors))
    linework = np.concatenate((lines, outer_rings, inner_rings))

    coordinates, line_index = shapely.get_coordinates(linework, return_index=True)
    segments = np.stack((coordinates[:-1, :], coordinates[1:]), axis=1)
    keep = np.diff(line_index) == 0
    segments = segments[keep]
    segment_to_line = line_index[1:][keep]
    segment_linestrings = shapely.linestrings(
        segments.reshape((-1, 2)), indices=np.repeat(np.arange(len(segments)), 2)
    )

    # We only use the STRtree to check bounding boxes. It does return the
    # intersection result, even if we specify the predicate.
    rtree = shapely.STRtree(segment_linestrings)
    i, j = rtree.query(segment_linestrings)

    # Compute intersections
    # Skip self-intersections
    keep = segment_to_line[i] != segment_to_line[j]
    i = i[keep]
    j = j[keep]
    i, j, intersections = compute_intersections(segment_linestrings, i, j)
    segments_i = segments[i]
    segments_j = segments[j]

    # Find minimum squared distance to vertices of intersecting edges. If the
    # minimum distance is (approximately) zero, the intersection is represented
    # by a vertex already.
    distance_i = np.linalg.norm(
        intersections[:, np.newaxis, :] - segments_i, axis=2
    ).min(axis=1)
    distance_j = np.linalg.norm(
        intersections[:, np.newaxis, :] - segments_j, axis=2
    ).min(axis=1)
    unresolved = (distance_i > tolerance) | (distance_j > tolerance)
    return intersections[unresolved]


def check_linework(
    polygons: gpd.GeoSeries,
    lines: gpd.GeoSeries,
    on_unresolved_intersection: str,
) -> None:
    intersections = linework_intersections(
        polygons.geometry,
        lines.geometry,
        tolerance=1.0e-6,
    )
    n_intersection = len(intersections)
    if n_intersection > 0:
        msg = (
            f"{n_intersection} unresolved intersections between polygon "
            "boundary or line segments.\nRun "
            "pandamesh.find_edge_intersections(gdf.geometry) to identify the "
            "intersection locations.\nIntersections can be resolved using "
            "the pandamesh.Preprocessor."
        )
        if on_unresolved_intersection == "error":
            raise ValueError(msg)
        elif on_unresolved_intersection == "warn":
            warnings.warn(msg)
    return


def find_edge_intersections(geometry: gpd.Geoseries) -> gpd.GeoSeries:
    """
    Find all unresolved intersections between polygon boundaries, linestring,
    and linearring edges.

    A "resolved" intersection is one where the intersection of two lines is
    represented by a vertex in both lines. Unresolved means: an intersection
    which is not represented by an explicit vertex in the geometries.

    Parameters
    ----------
    geometry: gpd.GeoSeries
        Points, lines, polygons.

    Returns
    -------
    intersections: gpd.GeoSeries
        Locations (points) of intersections.
    """
    if not isinstance(geometry, gpd.GeoSeries):
        raise TypeError(
            f"Expected geopandas.GeoSeries, received: {type(geometry).__name__}"
        )
    polygons, lines, _ = separate_geometry(geometry)
    intersections = linework_intersections(
        polygons,
        lines,
        tolerance=1.0e-6,
    )
    return gpd.GeoSeries(gpd.points_from_xy(*intersections.T))


def check_points(
    points: gpd.GeoSeries,
    polygons: gpd.GeoSeries,
) -> None:
    """Check whether points are contained by a polygon."""
    within = gpd.GeoDataFrame(geometry=points).sjoin(
        df=gpd.GeoDataFrame(geometry=polygons),
        predicate="within",
    )
    n_outside = len(points) - len(within)
    if n_outside != 0:
        raise ValueError(f"{n_outside} points detected outside of a polygon")
    return


def _separate(
    geometry: GeometryArray,
) -> Tuple[BoolArray, BoolArray, BoolArray]:
    geometry_id = shapely.get_type_id(geometry)
    allowed_types = (POINT, LINESTRING, LINEARRING, POLYGON)
    if not np.isin(geometry_id, allowed_types).all():
        received = ", ".join(
            [GEOM_NAMES[geom_id] for geom_id in np.unique(geometry_id)]
        )
        raise TypeError(
            "GeoDataFrame contains unsupported geometry types. Geometry "
            "should be one of Point, LineString, LinearRing, and Polygon "
            f"geometries. Received: {received}\n"
            "Call geopandas.GeoDataFrame.explode() to explode multi-part "
            "geometries into multiple single geometries."
        )
    return (
        geometry_id == POLYGON,
        (geometry_id == LINESTRING) | (geometry_id == LINEARRING),
        geometry_id == POINT,
    )


def separate_geometry(
    geometry: GeometryArray,
) -> Tuple[GeometryArray, GeometryArray, GeometryArray]:
    is_polygon, is_line, is_point = _separate(geometry)
    return (
        geometry[is_polygon],
        geometry[is_line],
        geometry[is_point],
    )


def separate_geodataframe(
    gdf: gpd.GeoDataFrame,
    intersecting_edges: str,
) -> Tuple[gpd.GeoDataFrame, gpd.GeoDataFrame, gpd.GeoDataFrame]:
    acceptable = ("error", "warn", "ignore")
    if intersecting_edges not in acceptable:
        raise ValueError(
            f"intersecting_edges should be one of {acceptable}, "
            f"received: {intersecting_edges}"
        )

    is_polygon, is_line, is_point = _separate(gdf["geometry"])
    polygons = gdf.loc[is_polygon].copy()
    lines = gdf.loc[is_line].copy()
    points = gdf.loc[is_point].copy()

    if len(polygons) == 0:
        raise ValueError("No polygons provided")

    for df in (polygons, lines, points):
        df["cellsize"] = df["cellsize"].astype(float)
        df.crs = None

    check_polygons(polygons.geometry)
    check_lines(lines.geometry)
    if intersecting_edges != "ignore":
        check_linework(polygons.geometry, lines.geometry, intersecting_edges)
    check_points(points.geometry, polygons.geometry)

    return polygons, lines, points


def move_origin(
    gdf: gpd.GeoDataFrame,
    xoff: float,
    yoff: float,
):
    if xoff == 0.0 and yoff == 0.0:
        return gdf.copy()
    else:
        moved = gdf.copy()
        moved["geometry"] = gdf["geometry"].translate(xoff=-xoff, yoff=-yoff)
    return moved


def central_origin(
    gdf: gpd.GeoDataFrame, shift_origin: bool
) -> Tuple[gpd.GeoDataFrame, float, float]:
    if shift_origin:
        xmin, ymin, xmax, ymax = gdf.total_bounds
        xoff = 0.5 * (xmin + xmax)
        yoff = 0.5 * (ymin + ymax)
    else:
        xoff = 0.0
        yoff = 0.0
    gdf = move_origin(gdf, xoff, yoff)
    return gdf, xoff, yoff


def to_ugrid(vertices: FloatArray, faces: IntArray) -> "xugrid.Ugrid2d":  # type: ignore # noqa pragma: no cover
    try:
        import xugrid
    except ImportError:
        raise ImportError(
            "xugrid must be installed to return generated result a xugrid.Ugrid2d"
        )
    return xugrid.Ugrid2d(*vertices.T, -1, faces)


def to_geodataframe(vertices: FloatArray, faces: IntArray) -> gpd.GeoDataFrame:
    n_face, n_vertex = faces.shape
    if n_vertex == 3:  # no fill values
        coordinates = vertices[faces]
        geometry = shapely.polygons(coordinates)
    else:  # Possible fill values (-1)
        # Group them by number of vertices.
        valid = faces >= 0
        n_valid = valid.sum(axis=1)
        grouped_by_n_vertex = pd.DataFrame(
            {"i": np.arange(n_face), "n": n_valid}
        ).groupby("n")
        geometries = []
        for n_vertex, group in grouped_by_n_vertex:
            group_faces = faces[group["i"], :n_vertex]
            group_coordinates = vertices[group_faces]
            geometries.append(shapely.polygons(group_coordinates))
        geometry = np.concatenate(geometries)
    return gpd.GeoDataFrame(geometry=geometry)


class Grouper:
    """
    Wrapper around pd.DataFrame().groupby().

    Group by ``a_index``, then iterates over the groups, returning a single
    value of ``a`` and potentially multiple values for ``b``.
    """

    def __init__(
        self,
        a,
        a_index,
        b,
        b_index,
    ):
        self.a = np.asarray(a)
        self.b = np.asarray(b)
        self.grouped = iter(
            pd.DataFrame(data={"a": a_index, "b": b_index}).groupby("a")
        )

    def __iter__(self):
        return self

    def __next__(self):
        try:
            a_index, group = next(self.grouped)
            b_index = group["b"]
            return self.a[a_index], self.b[b_index]
        except StopIteration:
            raise StopIteration



import geopandas as gpd
import numpy as np
import pytest
import shapely.geometry as sg

from pandamesh import gmsh_geometry as gg
from pandamesh.gmsh_mesher import gmsh_env


class TestGmshGeometry:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.outer_coords = np.array(
            [(0.0, 0.0), (10.0, 0.0), (10.0, 10.0), (0.0, 10.0)]
        )
        self.inner_coords = np.array([(3.0, 3.0), (7.0, 3.0), (7.0, 7.0), (3.0, 7.0)])
        self.line_coords = np.array([(2.0, 8.0), (8.0, 2.0)])
        self.line = sg.LineString(self.line_coords)
        self.polygon = sg.Polygon(self.outer_coords)
        self.donut = sg.Polygon(self.outer_coords, holes=[self.inner_coords])
        self.refined = sg.Polygon(self.inner_coords)
        y = np.arange(0.5, 10.0, 0.5)
        x = np.full(y.size, 1.0)
        self.points_embed = gpd.points_from_xy(x, y)
        self.line_embed = sg.LineString(
            [
                [9.0, 2.0],
                [9.0, 8.0],
            ]
        )
        self.polygons = gpd.GeoDataFrame(
            {"cellsize": [1.0], "__polygon_id": [1]}, geometry=[self.donut]
        )

    def test_polygon_info(self):
        info, vertices, cellsizes, index = gg.polygon_info(self.polygon, 1.0, 0, 0)
        expected_info = gg.PolygonInfo(0, 4, [], [], 0)
        assert np.allclose(vertices, self.outer_coords)
        assert info == expected_info
        assert np.allclose(cellsizes, 1.0)
        assert index == 4

        info, vertices, cellsizes, index = gg.polygon_info(self.donut, 1.0, 0, 0)
        expected_info = gg.PolygonInfo(0, 4, [4], [4], 0)
        assert np.allclose(vertices, [self.outer_coords, self.inner_coords])
        assert info == expected_info
        assert np.allclose(cellsizes, 1.0)
        assert index == 8

    def test_linestring_info(self):
        info, vertices, cellsizes, index = gg.linestring_info(self.line, 1.0, 0, 0)
        expected_info = gg.LineStringInfo(0, 2, 0)
        assert np.allclose(vertices, self.line_coords)
        assert info == expected_info
        assert np.allclose(cellsizes, 1.0)
        assert index == 2

    def test_add_vertices(self):
        with gmsh_env():
            gg.add_vertices(
                [(0.0, 0.0), (1.0, 1.0)],
                [1.0, 1.0],
                [1, 2],
            )

    def test_add_linestrings(self):
        info = gg.LineStringInfo(0, 2, 0)
        with gmsh_env():
            gg.add_vertices(
                [(0.0, 0.0), (1.0, 1.0)],
                [1.0, 1.0],
                [0, 1],
            )
            gg.add_linestrings([info], [0, 1])

    def test_add_curve_loop(self):
        with gmsh_env():
            gg.add_vertices(
                [(0.0, 0.0), (1.0, 0.0), (1.0, 1.0), (0.0, 1.0)],
                [1.0, 1.0, 1.0, 1.0],
                [0, 1, 2, 3],
            )
            curve_loop_tag = gg.add_curve_loop([0, 1, 2, 3])
        assert curve_loop_tag == 1

    def test_add_polygons(self):
        info = gg.PolygonInfo(0, 4, [4], [4], 0)
        vertices = np.vstack([self.inner_coords, self.outer_coords])
        cellsizes = np.full(vertices.size, 1.0)
        tags = np.arange(vertices.size)
        with gmsh_env():
            gg.add_vertices(vertices, cellsizes, tags)
            loop_tags, plane_tags = gg.add_polygons([info], tags)
        assert loop_tags == [1, 2]
        assert plane_tags == [0]

    def test_add_points(self):
        x = np.arange(0.5, 10.0, 0.5)
        y = np.full(x.size, 1.0)
        ids = np.arange(x.size) + 1
        gdf = gpd.GeoDataFrame(geometry=gpd.points_from_xy(x, y))
        gdf["__polygon_id"] = ids
        gdf["cellsize"] = 1.0
        with gmsh_env():
            indices, embedded_in = gg.add_points(gdf)
        assert np.array_equal(indices, ids)
        assert np.array_equal(embedded_in, ids)

    def test_collect_polygons(self):
        gdf = gpd.GeoDataFrame(geometry=[self.polygon])
        gdf["cellsize"] = 1.0
        gdf["__polygon_id"] = 1
        index, vertices, cellsizes, features = gg.collect_polygons(gdf, 0)
        assert index == 4
        assert np.allclose(vertices, self.outer_coords)
        assert np.allclose(cellsizes, 1.0)
        assert features == [gg.PolygonInfo(0, 4, [], [], 1)]

    def test_collect_linestrings(self):
        gdf = gpd.GeoDataFrame(geometry=[self.line])
        gdf["cellsize"] = 1.0
        gdf["__polygon_id"] = 1
        index, vertices, cellsizes, features = gg.collect_linestrings(gdf, 0)
        assert index == 2
        assert np.allclose(vertices, self.line_coords)
        assert np.allclose(cellsizes, 1.0)
        assert features == [gg.LineStringInfo(0, 2, 1)]

    def test_collect_points(self):
        x = np.arange(0.5, 10.0, 0.5)
        y = np.full(x.size, 1.0)
        gdf = gpd.GeoDataFrame(geometry=gpd.points_from_xy(x, y))
        xy = gg.collect_points(gdf)
        assert np.allclose(xy, np.column_stack([x, y]))

    def test_embed_where_linestring(self):
        line_gdf = gpd.GeoDataFrame({"cellsize": [0.5]}, geometry=[self.line_embed])
        actual = gg.embed_where(line_gdf, self.polygons)
        assert np.allclose(actual["cellsize"], 0.5)
        assert np.allclose(actual["__polygon_id"], 1)
        assert actual.geometry.iloc[0] == self.line_embed

    def test_embed_where_points(self):
        points_gdf = gpd.GeoDataFrame(geometry=self.points_embed)
        points_gdf["cellsize"] = 0.25
        actual = gg.embed_where(points_gdf, self.polygons)
        assert np.allclose(actual["cellsize"], 0.25)
        assert np.allclose(actual["__polygon_id"], 1)
        assert (actual.geometry.to_numpy() == self.points_embed).all()

    def test_add_geometry(self):
        line_gdf = gpd.GeoDataFrame({"cellsize": [0.5]}, geometry=[self.line_embed])
        points_gdf = gpd.GeoDataFrame(geometry=self.points_embed)
        points_gdf["cellsize"] = 0.25
        with gmsh_env():
            gg.add_geometry(self.polygons, line_gdf, points_gdf)

    def test_add_distance_points(self):
        with gmsh_env():
            indices = gg.add_distance_points(self.points_embed)
        assert np.array_equal(indices, np.arange(1, 20))

    def test_add_distance_linestring(self):
        with gmsh_env():
            indices = gg.add_distance_linestring(self.line, distance=20)
        assert np.array_equal(indices, [1])

        with gmsh_env():
            indices = gg.add_distance_linestring(self.line, distance=5)
        assert np.array_equal(indices, [1, 2])

        with gmsh_env():
            indices = gg.add_distance_linestring(self.line, distance=3)
        assert np.array_equal(indices, [1, 2, 3])

    def test_add_distance_linestrings(self):
        lines = gpd.GeoSeries([self.line, self.line_embed])
        with gmsh_env():
            indices = gg.add_distance_linestrings(lines, spacing=np.array([20, 20]))
        assert np.array_equal(indices, [1, 2])

        with gmsh_env():
            indices = gg.add_distance_linestrings(lines, spacing=np.array([3, 3]))
        assert np.array_equal(indices, [1, 2, 3, 4, 5])

        with gmsh_env():
            indices = gg.add_distance_linestrings(lines, spacing=np.array([20, 3]))
        assert np.array_equal(indices, [1, 2, 3])

    def test_add_distance_polygons(self):
        polygons = gpd.GeoSeries([self.donut, self.refined])
        with gmsh_env():
            indices = gg.add_distance_polygons(polygons, spacing=np.array([100, 100]))
        # Exterior, interior, exterior
        assert np.array_equal(indices, [1, 2, 3])

        with gmsh_env():
            indices = gg.add_distance_polygons(polygons, spacing=np.array([3, 3]))
        # Exterior, interior, exterior
        assert np.array_equal(indices, np.arange(1, 27))

    def test_add_distance_geometry(self):
        geometry = gpd.GeoSeries([sg.MultiPolygon([self.donut, self.refined])])
        with pytest.raises(TypeError, match="Geometry should be one of"):
            gg.add_distance_geometry(geometry, np.array([1.0]))

        # One point on each edge
        geometry = gpd.GeoSeries([self.polygon])
        with gmsh_env():
            indices = gg.add_distance_geometry(geometry, np.array([10.0]))
        assert np.array_equal(indices, [1, 2, 3, 4])

        # Three points
        geometry = gpd.GeoSeries([self.line])
        with gmsh_env():
            indices = gg.add_distance_geometry(geometry, np.array([3.0]))
        assert np.array_equal(indices, [1, 2, 3])

        # Twenty points, one to one
        geometry = gpd.GeoSeries(self.points_embed)
        with gmsh_env():
            indices = gg.add_distance_geometry(geometry, np.array([np.nan]))
        assert np.array_equal(indices, np.arange(1, 20))

        # All together
        geometry = gpd.GeoSeries(
            np.concatenate([[self.polygon], [self.line], self.points_embed])
        )
        with gmsh_env():
            indices = gg.add_distance_geometry(geometry, np.array([10.0, 3.0, np.nan]))
        assert np.array_equal(indices, np.arange(1, 25))



from pandamesh.enum_base import FlexibleEnum


class MeshAlgorithm(FlexibleEnum):
    """
    Gmsh meshing algorithm. Each algorithm has its own advantages and
    disadvantages.

    For all 2D unstructured algorithms a Delaunay mesh that contains all
    the points of the 1D mesh is initially constructed using a
    divide-and-conquer algorithm. Missing edges are recovered using edge
    swaps. After this initial step several algorithms can be applied to
    generate the final mesh:

    * The MeshAdapt algorithm is based on local mesh modifications. This
      technique makes use of edge swaps, splits, and collapses: long edges
      are split, short edges are collapsed, and edges are swapped if a
      better geometrical configuration is obtained.
    * The Delaunay algorithm is inspired by the work of the GAMMA team at
      INRIA. New points are inserted sequentially at the circumcenter of
      the element that has the largest adimensional circumradius. The mesh
      is then reconnected using an anisotropic Delaunay criterion.
    * The Frontal-Delaunay algorithm is inspired by the work of S. Rebay.
    * Other experimental algorithms with specific features are also
      available. In particular, Frontal-Delaunay for Quads is a variant of
      the Frontal-Delaunay algorithm aiming at generating right-angle
      triangles suitable for recombination; and BAMG allows to generate
      anisotropic triangulations.

    For very complex curved surfaces the MeshAdapt algorithm is the most robust.
    When high element quality is important, the Frontal-Delaunay algorithm should
    be tried. For very large meshes of plane surfaces the Delaunay algorithm is
    the fastest; it usually also handles complex mesh size fields better than the
    Frontal-Delaunay. When the Delaunay or Frontal-Delaunay algorithms fail,
    MeshAdapt is automatically triggered. The Automatic algorithm uses
    Delaunay for plane surfaces and MeshAdapt for all other surfaces.
    """

    MESH_ADAPT = 1
    """
    Local mesh modifications using edge swaps, splits, and collapses. Robust
    for complex curved surfaces.
    """

    AUTOMATIC = 2
    """
    Uses Delaunay for plane surfaces and MeshAdapt for all other surfaces.
    """

    INITIAL_MESH_ONLY = 3
    """Generates only the initial Delaunay triangulation."""

    FRONTAL_DELAUNAY = 5
    """Good for high element quality."""

    BAMG = 7
    """Experimental algorithm for generating anisotropic triangulations."""

    FRONTAL_DELAUNAY_FOR_QUADS = 8
    """
    Variant of Frontal-Delaunay aiming to generate right-angle triangles
    suitable for recombination.
    """

    PACKING_OF_PARALLELLOGRAMS = 9
    """Experimental algorithm for parallelogram-based mesh generation."""

    QUASI_STRUCTURED_QUAD = 11
    """
    Combines an initial unstructured quad mesh with topological improvements
    guided by cross fields to produce quasi-structured meshes with few
    irregular vertices.
    """


class SubdivisionAlgorithm(FlexibleEnum):
    """
    Controls how Gmsh recombines triangles to form quads.

    The default recombination algorithm might leave some triangles in the mesh,
    if recombining all the triangles leads to badly shaped quads. In such
    cases, to generate full-quad meshes, you can either subdivide the resulting
    hybrid mesh (ALL_QUADRANGLES), or use the full-quad recombination
    algorithm, which will automatically perform a coarser mesh followed by
    recombination, smoothing and subdivision.
    """

    NONE = 0
    """
    No subdivision is applied. The mesh remains as is after the initial
    recombination process, potentially leaving some triangles in the mesh.
    """
    ALL_QUADRANGLES = 1
    """
    Subdivides the mesh to convert all elements into quadrangles. This method
    ensures a full-quad mesh by subdividing any remaining triangles after the
    initial recombination process.
    """
    BARYCENTRIC = 3
    """
    Applies barycentric subdivision to the mesh. This method subdivides each
    element by connecting its barycenter to the midpoints of its edges,
    resulting in a refined mesh with increased element count.
    """


class FieldCombination(FlexibleEnum):
    """
    Controls how cell size fields are combined in Gmsh when they are found at
    the same location.
    """

    MIN = "Min"
    """Use the minimum size."""
    MAX = "Max"
    """Use the maximum size."""


class GeneralVerbosity(FlexibleEnum):
    """Gmsh level of information printed."""

    SILENT = 0
    """No output is printed. All messages are suppressed."""

    ERRORS = 1
    """
    Only error messages are printed. This level is useful when you want to be
    alerted only to critical issues that prevent correct execution.
    """

    WARNINGS = 2
    """
    Error and warning messages are printed. This level adds important
    cautionary information that doesn't necessarily prevent execution but might
    affect results.
    """

    DIRECT = 3
    """
    Errors, warnings, and direct output from Gmsh commands are printed. This
    level is useful for seeing immediate results of operations without too much
    extra information.
    """

    INFORMATION = 4
    """
    Errors, warnings, direct output, and additional informational messages are
    printed. This level provides more detailed feedback about the progress and
    state of operations.
    """

    STATUS = 5
    """
    Errors, warnings, direct output, information, and status updates are
    printed. This level offers comprehensive feedback, including progress
    indicators for longer operations.
    """

    DEBUG = 99
    """
    All possible output is printed, including detailed debugging information.
    This level is extremely verbose and is typically used for troubleshooting
    or development purposes.
    """



import re

import geopandas as gpd
import numpy as np
import pytest
import shapely
import shapely.geometry as sg

from pandamesh import common

a = sg.Polygon(
    [
        (0.0, 0.0),
        (1.0, 0.0),
        (1.0, 1.0),
        (0.0, 1.0),
    ]
)
b = sg.Polygon(
    [
        (0.5, 0.0),
        (1.5, 0.0),
        (1.5, 1.0),
        (0.5, 1.0),
    ]
)
c = sg.Polygon(
    [
        (1.0, 0.0),
        (2.0, 0.0),
        (2.0, 1.0),
        (1.0, 1.0),
    ]
)
d = sg.Polygon(
    [
        (2.0, 0.0),
        (3.0, 0.0),
        (3.0, 1.0),
        (2.0, 1.0),
    ]
)
# Bowtie:
e = sg.Polygon(
    [
        (3.0, 0.0),
        (4.0, 1.0),
        (3.0, 1.0),
        (4.0, 0.0),
    ]
)

La = sg.LineString(
    [
        (0.25, 0.25),
        (0.75, 0.75),
    ]
)
Lb = sg.LineString(
    [
        (0.25, 0.75),
        (0.75, 0.25),
    ]
)
Lc = sg.LineString(
    [
        (2.25, 0.25),
        (2.75, 0.75),
    ]
)
Ld = sg.LineString(
    [
        (0.6, 0.5),
        (1.5, 0.5),
    ]
)
# Bowtie
Le = sg.LineString(
    [
        (3.0, 0.0),
        (4.0, 1.0),
        (3.0, 1.0),
        (4.0, 0.0),
    ]
)
# Outside
Lf = sg.LineString(
    [
        (1.0, 2.0),
        (2.0, 2.0),
    ]
)

Ra = sg.LinearRing(
    [
        (0.0, 0.0),
        (1.0, 0.0),
        (1.0, 1.0),
        (0.0, 1.0),
    ]
)


pa = sg.Point(0.5, 0.5)
pb = sg.Point(0.5, 1.5)


def test_flatten():
    assert common.flatten([[]]) == []
    assert common.flatten([[1]]) == [1]
    assert common.flatten([[1], [2, 3]]) == [1, 2, 3]


def test_flatten_geometry():
    point = sg.Point(0, 0)
    assert common.flatten_geometry(point) == [point]

    multi_point = sg.MultiPoint([sg.Point(0, 0), sg.Point(1, 1)])
    assert common.flatten_geometry(multi_point) == list(multi_point.geoms)

    point = sg.Point(0, 0)
    line = sg.LineString([(1, 1), (2, 2)])
    polygon = sg.Polygon([(3, 3), (4, 4), (4, 3)])
    multi_point = sg.MultiPoint([sg.Point(5, 5), sg.Point(6, 6)])

    collection = sg.GeometryCollection([point, line, polygon, multi_point])
    expected = [point, line, polygon, sg.Point(5, 5), sg.Point(6, 6)]
    assert common.flatten_geometry(collection) == expected

    inner_collection = sg.GeometryCollection([point, multi_point])
    nested_collection = sg.GeometryCollection([inner_collection, line, polygon])
    expected = [point, sg.Point(5, 5), sg.Point(6, 6), line, polygon]
    assert common.flatten_geometry(nested_collection) == expected

    # Also test flatten_geometries
    geometries = [
        point,
        sg.GeometryCollection([multi_point]),
        sg.GeometryCollection([line, polygon]),
    ]
    actual = common.flatten_geometries(geometries)
    assert isinstance(actual, np.ndarray)
    assert all(actual == expected)


def test_check_geodataframe():
    with pytest.raises(TypeError, match="Expected GeoDataFrame"):
        common.check_geodataframe([1, 2, 3], {})

    gdf = gpd.GeoDataFrame(geometry=[pa, pb])
    with pytest.raises(
        ValueError,
        match=re.escape("These column(s) are required but are missing: cellsize"),
    ):
        common.check_geodataframe(gdf, {"cellsize"})

    gdf["cellsize"] = 1.0
    gdf.index = [0, "1"]
    with pytest.raises(ValueError, match="geodataframe index is not integer typed"):
        common.check_geodataframe(gdf, {"cellsize"}, check_index=True)

    empty = gdf.loc[[]]
    with pytest.raises(ValueError, match="Dataframe is empty"):
        common.check_geodataframe(empty, {"cellsize"})

    gdf.index = [0, 0]
    with pytest.raises(ValueError, match="geodataframe index contains duplicates"):
        common.check_geodataframe(gdf, {"cellsize"}, check_index=True)

    gdf.index = [0, 1]
    common.check_geodataframe(gdf, {"cellsize"}, check_index=True)


def test_intersecting_features():
    polygons = gpd.GeoSeries(data=[a, b, c, d], index=[0, 1, 2, 3])
    ia, ib = common.intersecting_features(polygons, "polygon")
    assert np.array_equal(ia, [0, 1])
    assert np.array_equal(ib, [1, 2])

    linestrings = gpd.GeoSeries(data=[La, Lb, Lc], index=[0, 1, 2])
    ia, ib = common.intersecting_features(linestrings, "linestring")
    assert np.array_equal(ia, [0])
    assert np.array_equal(ib, [1])


def test_check_lines():
    # Complex (self-intersecting) linestring
    linestrings = gpd.GeoSeries(data=[La, Lb, Lc, Le], index=[0, 1, 2, 3])
    with pytest.raises(ValueError, match="1 cases of complex lines detected"):
        common.check_lines(linestrings)

    # Valid input
    linestrings = gpd.GeoSeries(data=[La, Lc], index=[0, 1])
    common.check_lines(linestrings)


def test_compute_intersections():
    segments = shapely.linestrings(
        [
            [
                [0.0, 5.0],
                [10.0, 5.0],
            ],
            [
                [5.0, 0.0],
                [5.0, 10.0],
            ],
            [
                [0.0, -5.0],
                [10.0, -5.0],
            ],
        ]
    )
    i = np.array([0, 0, 1, 1, 2])
    j = np.array([0, 1, 1, 2, 2])

    _i, _j, actual = common.compute_intersections(segments, i, j)
    assert np.array_equal(_i, [0])
    assert np.array_equal(_j, [1])
    assert np.array_equal(actual, [[5.0, 5.0]])

    # No intersections
    _i, _j, actual = common.compute_intersections(segments, i, i)
    assert np.array_equal(_i, [])
    assert np.array_equal(_j, [])
    assert np.array_equal(actual, np.array([]).reshape((-1, 2)))

    i = np.array([0])
    j = np.array([2])
    _i, _j, actual = common.compute_intersections(segments, i, j)
    assert np.array_equal(_i, [])
    assert np.array_equal(_j, [])
    assert np.array_equal(actual, np.array([]).reshape((-1, 2)))

    # Parallel
    segments = shapely.linestrings(
        [
            [
                [0.0, 5.0],
                [10.0, 5.0],
            ],
            [
                [3.0, 5.0],
                [13.0, 5.0],
            ],
        ]
    )
    i = np.array([0])
    j = np.array([1])
    _i, _j, actual = common.compute_intersections(segments, i, j)
    assert np.array_equal(_i, [0, 0])
    assert np.array_equal(_j, [1, 1])
    expected = np.array(
        [
            [3.0, 5.0],
            [10.0, 5.0],
        ]
    )
    assert np.array_equal(actual, expected)


def test_check_polygons():
    polygons = gpd.GeoSeries(data=[a, b, c, d, e], index=[0, 1, 2, 3, 4])
    with pytest.raises(ValueError, match="1 cases of complex polygon detected"):
        common.check_polygons(polygons)

    polygons = gpd.GeoSeries(data=[a, b, c, d], index=[0, 1, 2, 3])
    with pytest.raises(ValueError, match="2 cases of intersecting polygon detected"):
        common.check_polygons(polygons)


def test_check_points():
    points = gpd.GeoSeries(data=[pa, pb], index=[0, 1])
    polygons = gpd.GeoSeries(data=[a, b, c, d, e], index=[0, 1, 2, 3, 4])
    with pytest.raises(ValueError, match="1 points detected outside"):
        common.check_points(points, polygons)


def test_separate_geometry():
    bad = np.array([sg.MultiPolygon([a, d])])
    with pytest.raises(
        TypeError, match="GeoDataFrame contains unsupported geometry types"
    ):
        common.separate_geometry(bad)

    geometry = np.array([a, c, La, Lc, pa, Ra, d])
    polygons, lines, points = common.separate_geometry(geometry)
    assert all(polygons == [a, c, d])
    assert all(lines == [La, Lc, Ra])
    assert all(points == [pa])


def test_separate_geodataframe():
    gdf = gpd.GeoDataFrame(geometry=[a, c, d, La, Lc, pa])
    gdf["cellsize"] = 1.0

    with pytest.raises(ValueError, match="intersecting_edges should be one of"):
        common.separate_geodataframe(gdf, intersecting_edges="abc")

    polygons, linestrings, points = common.separate_geodataframe(
        gdf, intersecting_edges="error"
    )
    assert isinstance(polygons.geometry.iloc[0], sg.Polygon)
    assert isinstance(linestrings.geometry.iloc[0], sg.LineString)
    assert isinstance(points.geometry.iloc[0], sg.Point)

    # Make sure it works for single elements
    gdf = gpd.GeoDataFrame(geometry=[a])
    gdf["cellsize"] = 1.0
    common.separate_geodataframe(gdf, intersecting_edges="error")

    gdf = gpd.GeoDataFrame(geometry=[a, La])
    gdf["cellsize"] = 1.0
    polygons, linestrings, points = common.separate_geodataframe(
        gdf, intersecting_edges="error"
    )

    # Make sure cellsize is cast to float
    gdf = gpd.GeoDataFrame(geometry=[a, La])
    gdf["cellsize"] = "1"
    dfs = common.separate_geodataframe(gdf, intersecting_edges="error")
    for df in dfs:
        assert np.issubdtype(df["cellsize"].dtype, np.floating)

    with pytest.raises(
        TypeError, match="GeoDataFrame contains unsupported geometry types"
    ):
        gdf = gpd.GeoDataFrame(geometry=[sg.MultiPolygon([a, b])])
        common.separate_geodataframe(gdf, intersecting_edges="error")


def test_central_origin():
    gdf = gpd.GeoDataFrame(geometry=[d])
    back, x, y = common.central_origin(gdf, False)
    assert gdf is not back
    assert x == 0
    assert y == 0

    back, x, y = common.central_origin(gdf, True)
    assert np.allclose(x, 2.5)
    assert np.allclose(y, 0.5)
    assert np.array_equal(back.total_bounds, [-0.5, -0.5, 0.5, 0.5])


def test_grouper():
    a = np.array([1, 2, 3, 4])
    a_index = np.array([0, 0, 1, 1])
    b = np.array([10, 20, 30, 40, 50])
    b_index = np.array([0, 1, 2, 3])
    with pytest.raises(ValueError, match="All arrays must be of the same length"):
        grouper = common.Grouper(a, [0, 0, 1], b, b_index)

    grouper = common.Grouper(a, a_index, b, b_index)

    # First group
    a_val, b_vals = next(grouper)
    assert a_val == 1
    assert np.array_equal(b_vals, np.array([10, 20]))

    # Second group
    a_val, b_vals = next(grouper)
    assert a_val == 2
    assert np.array_equal(b_vals, np.array([30, 40]))

    # As iterator
    grouper = common.Grouper(a, a_index, b, b_index)
    results = list(grouper)
    assert results[0][0] == 1
    assert np.array_equal(results[0][1], np.array([10, 20]))
    assert results[1][0] == 2
    assert np.array_equal(results[1][1], np.array([30, 40]))


def test_to_geodataframe():
    # Two triangles
    vertices = np.array([[0, 0], [1, 0], [0, 1], [1, 1]])
    faces = np.array([[0, 1, 2], [1, 3, 2]])

    gdf = common.to_geodataframe(vertices, faces)
    assert isinstance(gdf, gpd.GeoDataFrame)
    assert len(gdf == 2)

    # Check the coordinates of the first triangle
    assert np.allclose(
        np.array(gdf.geometry[0].exterior.coords),
        np.array([(0, 0), (1, 0), (0, 1), (0, 0)]),
    )

    # Check the coordinates of the second triangle
    assert np.allclose(
        np.array(gdf.geometry[1].exterior.coords),
        np.array([(1, 0), (1, 1), (0, 1), (1, 0)]),
    )

    # Define vertices
    vertices = np.array([[0, 0], [1, 0], [0, 1], [1, 1], [2, 0.5]], dtype=float)

    # Define faces (triangle: 0,1,2 and quadrangle: 1,3,4,2)
    faces = np.array([[0, 1, 2, -1], [1, 3, 4, 2]], dtype=int)

    # Call the function
    gdf = common.to_geodataframe(vertices, faces)

    # Assertions
    assert isinstance(gdf, gpd.GeoDataFrame)
    assert len(gdf) == 2  # Two polygons (one triangle, one quadrangle)


class TestLineworkIntersection:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.donut = sg.Polygon(
            [
                [0.0, 0.0],
                [10.0, 0.0],
                [10.0, 10.0],
                [0.0, 10.0],
            ],
            holes=[
                [
                    [2.0, 5.0],
                    [5.0, 8.0],
                    [8.0, 5.0],
                    [5.0, 2.0],
                ]
            ],
        )
        self.line0 = shapely.LineString(
            [
                [-2.0, 0.0],
                [12.0, 10.0],
            ]
        )
        self.line1 = shapely.LineString(
            [
                [5.5, 9.0],
                [9.0, 5.5],
            ]
        )
        self.gdf = gpd.GeoDataFrame(geometry=[self.donut, self.line0, self.line1])

    def test_find_edge_intersections(self):
        gdf = gpd.GeoDataFrame(geometry=[self.donut])
        with pytest.raises(TypeError, match="Expected geopandas.GeoSeries"):
            common.find_edge_intersections(gdf)

        actual = common.find_edge_intersections(gdf.geometry)
        assert len(actual) == 0

        gdf = gpd.GeoDataFrame(geometry=[self.donut, self.line0])
        actual = common.find_edge_intersections(gdf.geometry)
        assert len(actual) == 4

        gdf = gpd.GeoDataFrame(geometry=[self.donut, self.line1])
        actual = common.find_edge_intersections(gdf.geometry)
        assert len(actual) == 0

    def test_check_linework(self):
        gdf = gpd.GeoDataFrame(geometry=[self.donut, self.line1])
        polygons, lines, _ = common.separate_geometry(gdf.geometry)
        common.check_linework(polygons, lines, "error")

        gdf = gpd.GeoDataFrame(geometry=[self.donut, self.line0])
        polygons, lines, _ = common.separate_geometry(gdf.geometry)
        with pytest.raises(ValueError):
            common.check_linework(polygons, lines, "error")

        with pytest.warns(UserWarning):
            common.check_linework(polygons, lines, "warn")



import geopandas as gpd

import pandamesh as pm


def test_provinces_nl():
    gdf = pm.data.provinces_nl()
    assert isinstance(gdf, gpd.GeoDataFrame)


def test_south_america():
    gdf = pm.data.south_america()
    assert isinstance(gdf, gpd.GeoDataFrame)



import geopandas as gpd
import numpy as np
import pytest
import shapely
import shapely.geometry as sg

from pandamesh import preprocessor as pr

outer = sg.Polygon(
    [
        [0.0, 0.0],
        [10.0, 0.0],
        [10.0, 10.0],
        [0.0, 10.0],
    ]
)
inner = sg.Polygon(
    [
        [5.0, 2.0],
        [8.0, 5.0],
        [5.0, 8.0],
        [2.0, 5.0],
    ]
)

first = sg.Polygon(
    [
        [0.0, 0.0],
        [10.0, 0.0],
        [10.0, 10.0],
        [0.0, 10.0],
    ]
)
second = sg.Polygon(
    [
        [10.0, 2.0],
        [18.0, 2.0],
        [18.0, 8.0],
        [10.0, 8.0],
    ]
)
third = sg.Polygon(
    [
        [18.0, 2.0],
        [22.0, 2.0],
        [22.0, 8.0],
        [18.0, 8.0],
    ]
)

donut = sg.Polygon(
    [
        [0.0, 0.0],
        [10.0, 0.0],
        [10.0, 10.0],
        [0.0, 10.0],
    ],
    holes=[
        [
            [2.0, 5.0],
            [5.0, 8.0],
            [8.0, 5.0],
            [5.0, 2.0],
        ]
    ],
)

first_line = shapely.LineString(
    [
        [-5.0, 5.0],
        [6.0, 5.0],
    ]
)
second_line = shapely.LineString(
    [
        [5.0, -5.0],
        [5.0, 6.0],
    ]
)


def test_collect_exteriors():
    polygons = np.array([outer, donut])
    exteriors = pr.collect_exteriors(polygons)
    assert len(exteriors) == 2
    assert (shapely.get_type_id(exteriors) == 2).all()


def test_collect_interiors():
    polygons = np.array([donut, outer])
    interiors = pr.collect_interiors(polygons)
    assert len(interiors == 1)
    assert (shapely.get_type_id(interiors) == 2).all()


def test_filter_holes_and_assign_values():
    # Assume: two polygons. Union, then polygonized to create a partitions. Any
    # polygon found in a hole should disappear.
    polygons = np.array([donut, inner, second, third])
    # Overlap between the original outer and inner polygon.
    index_original = np.array([0, 1, 0, 1])
    index_union = np.array([0, 0, 1, 1])
    indexer = np.array([4, 3, 2, 1])

    filtered_polygons, filtered_indexer = pr.filter_holes_and_assign_values(
        polygons, index_original, index_union, indexer, ascending=True
    )
    assert all(filtered_polygons == [donut, inner])
    assert np.array_equal(filtered_indexer, [3, 3])

    filtered_polygons, filtered_indexer = pr.filter_holes_and_assign_values(
        polygons, index_original, index_union, indexer, ascending=False
    )
    assert all(filtered_polygons == [donut, inner])
    assert np.array_equal(filtered_indexer, [4, 4])


def test_locate_polygons():
    new = np.array([donut, inner])
    old = np.array([outer, inner])
    indexer = np.array([1, 0])
    polygons, indexer = pr.locate_polygons(new, old, indexer, True)
    assert all(polygons == [donut, inner])
    assert np.array_equal(indexer, [1, 0])

    polygons, indexer = pr.locate_polygons(new, old, indexer, False)
    assert all(polygons == [donut, inner])
    assert np.array_equal(indexer, [1, 1])


def test_locate_lines():
    old = np.array([first_line, second_line])
    new = np.array(
        [
            shapely.LineString([[-5.0, 5.0], [0.0, 5.0]]),
            shapely.LineString([[5.0, -5.0], [5.0, 0.0]]),
            shapely.LineString([[5.0, 0.0], [5.0, 5.0]]),
            shapely.LineString([[0.0, 5.0], [5.0, 5.0]]),
        ]
    )
    indexer = np.array([0, 1])
    located = pr.locate_lines(new, old, indexer)
    assert np.array_equal(located, [0, 1, 1, 0])


def test_merge_polyons():
    polygons = np.array([first, second])
    merged = pr.merge_polygons(polygons, None)
    assert len(merged == 1)
    assert (shapely.get_type_id(merged) == 3).all()

    polygons = np.array([inner, second])
    merged = pr.merge_polygons(polygons, None)
    assert len(merged == 2)
    assert (shapely.get_type_id(merged) == 3).all()


def test_preprocessor_init():
    geometry = [outer]
    values = [1.0, 2.0]

    with pytest.raises(ValueError, match="geometry and values shape mismatch"):
        pr.Preprocessor(geometry, values)

    point = sg.Point([1.0, 1.0])
    geometry = [outer, inner, second, first_line, point]
    p = pr.Preprocessor(geometry)
    assert p.values is None
    assert isinstance(p.polygons, np.ndarray)
    assert isinstance(p.lines, np.ndarray)
    assert isinstance(p.points, np.ndarray)
    assert np.array_equal(p.polygon_indexer, [0, 1, 2])
    assert np.array_equal(p.line_indexer, [3])
    assert np.array_equal(p.point_indexer, [4])

    values = [3.0, 2.0, 1.0, 0.5, 0.5]
    p = pr.Preprocessor(geometry, values)
    assert np.array_equal(p.values, [0.5, 1.0, 2.0, 3.0])
    assert np.array_equal(p.polygon_indexer, [3, 2, 1])
    assert np.array_equal(p.line_indexer, [0])
    assert np.array_equal(p.point_indexer, [0])


class TestPreprocessor:
    @pytest.fixture(autouse=True)
    def setup(self):
        points = [sg.Point([0.1, 0.1]), sg.Point([1.0, 1.0]), sg.Point([-5.0, 5.0])]
        self.geometry = [outer, inner, second, first_line, second_line, *points]
        self.values = [3.0, 1.0, 3.0, 0.5, 0.5, 1.0, 1.0, 1.0]
        self.p = pr.Preprocessor(self.geometry)
        self.vp = pr.Preprocessor(self.geometry, self.values)

    def test_copy_with(self):
        def test_value_equality(old, new):
            for k, v in new.items():
                assert v is old[k]

        p = self.p
        copied = p._copy_with()
        assert isinstance(copied, pr.Preprocessor)
        assert copied is not p
        test_value_equality(p.__dict__, copied.__dict__)

        copied = p._copy_with(values=[1, 2, 3, 4, 5, 6, 7])
        assert p.values is not copied.values
        new = copied.__dict__
        assert np.array_equal(new.pop("values"), [1, 2, 3, 4, 5, 6, 7])
        test_value_equality(p.__dict__, new)

    def test_to_geodataframe(self):
        gdf = self.p.to_geodataframe()
        assert isinstance(gdf, gpd.GeoDataFrame)
        assert set(gdf.columns) == {"geometry", "indexer"}
        assert np.array_equal(gdf["indexer"], [0, 1, 2, 3, 4, 5, 6, 7])

        gdf = self.vp.to_geodataframe()
        assert isinstance(gdf, gpd.GeoDataFrame)
        assert set(gdf.columns) == {"geometry", "indexer", "values"}
        assert np.array_equal(gdf["values"], self.values)

    def test_merge_polygons(self):
        merged = self.p.merge_polygons()
        assert isinstance(merged, pr.Preprocessor)
        assert len(merged.polygons == self.p.polygons)
        assert (shapely.get_type_id(merged.polygons) == 3).all()

        merged = self.vp.merge_polygons()
        assert len(merged.polygons) == 2
        assert np.array_equal(merged.polygon_indexer, [1, 2])
        assert (shapely.get_type_id(merged.polygons) == 3).all()

    def test_unify_polygons(self):
        unified = self.p.unify_polygons()
        assert isinstance(unified, pr.Preprocessor)
        assert len(unified.polygons) == 5
        assert (shapely.get_type_id(unified.polygons) == 3).all()
        assert np.array_equal(unified.polygon_indexer, [0, 0, 2, 0, 0])

        # Outer and second have same values, so they are merged.
        unified = self.vp.unify_polygons().merge_polygons()
        assert len(unified.polygons) == 2
        assert (shapely.get_type_id(unified.polygons) == 3).all()
        assert np.array_equal(unified.polygon_indexer, [1, 2])

    def test_clip_lines(self):
        clipped = self.p.clip_lines()
        assert isinstance(clipped, pr.Preprocessor)
        assert len(clipped.lines) == 2

        expected = [
            sg.LineString([[0.0, 5.0], [6.0, 5.0]]),
            sg.LineString([[5.0, 0.0], [5.0, 6.0]]),
        ]
        assert all(clipped.lines == expected)
        assert np.array_equal(clipped.line_indexer, [3, 4])

        p = pr.Preprocessor(geometry=[donut, first_line, second_line])
        clipped = p.clip_lines(distance=0.5)
        expected = [
            sg.LineString([[0.5, 5.0], [1.5, 5.0]]),
            sg.LineString([[5.0, 0.5], [5.0, 1.5]]),
        ]
        assert all(clipped.lines == expected)
        assert np.array_equal(clipped.line_indexer, [1, 2])

    def test_unify_lines(self):
        unified = self.p.unify_lines()
        assert isinstance(unified, pr.Preprocessor)
        assert len(unified.lines) == 4
        expected = [
            sg.LineString([[-5.0, 5.0], [5.0, 5.0]]),
            sg.LineString([[5.0, 5.0], [6.0, 5.0]]),
            sg.LineString([[5.0, -5.0], [5.0, 5.0]]),
            sg.LineString([[5.0, 5.0], [5.0, 6.0]]),
        ]
        assert all(unified.lines == expected)
        assert np.array_equal(unified.line_indexer, [3, 3, 4, 4])

    def test_clip_points(self):
        clipped = self.p.clip_points()
        assert isinstance(clipped, pr.Preprocessor)
        assert len(clipped.points) == 2
        assert np.array_equal(clipped.point_indexer, [5, 6])

        clipped = self.p.clip_points(distance=0.5)
        assert isinstance(clipped, pr.Preprocessor)
        assert len(clipped.points) == 1
        assert np.array_equal(clipped.point_indexer, [6])

    def test_interpolate_lines_to_points(self):
        with pytest.raises(ValueError, match="If values_as_distance is False"):
            self.p.interpolate_lines_to_points(distance=None, values_as_distance=False)

        interpolated = self.p.interpolate_lines_to_points(distance=1.0)
        assert isinstance(interpolated, pr.Preprocessor)
        assert len(interpolated.points) == 27
        assert np.array_equal(
            interpolated.point_indexer, [5, 6, 7] + [3] * 12 + [4] * 12
        )

        interpolated = self.vp.interpolate_lines_to_points(values_as_distance=True)
        assert isinstance(interpolated, pr.Preprocessor)
        assert len(interpolated.points) == 49
        assert np.array_equal(
            interpolated.point_indexer, [1, 1, 1] + [0] * 23 + [0] * 23
        )

    def test_snap_points(self):
        snapped = self.p.snap_points(0.05)
        assert isinstance(snapped, pr.Preprocessor)
        assert len(snapped.points) == 3
        assert np.array_equal(snapped.point_indexer, [5, 6, 7])

        snapped = self.p.snap_points(2.0)
        assert isinstance(snapped, pr.Preprocessor)
        assert len(snapped.points) == 2
        assert np.array_equal(snapped.point_indexer, [5, 7])

    def test_empty_ops(self):
        p = pr.Preprocessor(geometry=[outer])
        assert isinstance(p.clip_points(), pr.Preprocessor)
        assert isinstance(p.snap_points(1.0), pr.Preprocessor)
        assert isinstance(p.clip_lines(), pr.Preprocessor)
        assert isinstance(p.unify_lines(), pr.Preprocessor)
        assert isinstance(p.interpolate_lines_to_points(), pr.Preprocessor)
        assert isinstance(p.unify_polygons(), pr.Preprocessor)
        assert isinstance(p.merge_polygons(), pr.Preprocessor)
        assert isinstance(p.to_geodataframe(), gpd.GeoDataFrame)



from typing import List, NamedTuple, Tuple, Union

import geopandas as gpd
import numpy as np
import pandas as pd
import shapely
import shapely.geometry as sg

from pandamesh.common import FloatArray, IntArray, coord_dtype, gmsh, separate_geometry

Z_DEFAULT = 0.0
POINT_DIM = 0
LINE_DIM = 1
PLANE_DIM = 2


class PolygonInfo(NamedTuple):
    index: int
    size: int
    interior_indices: List[int]
    interior_sizes: List[int]
    polygon_id: int


class LineStringInfo(NamedTuple):
    index: int
    size: int
    embedded_in: Union[int, None]


def polygon_info(
    polygon: sg.Polygon, cellsize: float, index: int, polygon_id: int
) -> Tuple[PolygonInfo, FloatArray, FloatArray, int]:
    exterior_coords = np.array(polygon.exterior.coords)[:-1]
    size = len(exterior_coords)
    vertices = [exterior_coords]
    cellsizes = [np.full(size, cellsize)]
    info = PolygonInfo(index, size, [], [], polygon_id)
    index += size
    for interior in polygon.interiors:
        interior_coords = np.array(interior.coords)[:-1]
        vertices.append(interior_coords)
        size = len(interior_coords)
        cellsizes.append(np.full(size, cellsize))
        info.interior_indices.append(index)
        info.interior_sizes.append(size)
        index += size
    return info, vertices, cellsizes, index


def linestring_info(
    linestring: sg.LineString, cellsize: float, index: int, inside: Union[int, None]
) -> Tuple[LineStringInfo, FloatArray, FloatArray, int]:
    vertices = np.array(linestring.coords)
    size = len(vertices)
    cellsizes = np.full(size, cellsize)
    info = LineStringInfo(index, size, inside)
    index += size
    return info, vertices, cellsizes, index


def add_vertices(vertices, cellsizes, tags) -> None:
    for (x, y), cellsize, tag in zip(vertices, cellsizes, tags):
        gmsh.model.geo.addPoint(x, y, Z_DEFAULT, cellsize, tag)


def add_linestrings(
    features: List[LineStringInfo], tags: IntArray
) -> Tuple[IntArray, IntArray]:
    n_lines = sum(info.size - 1 for info in features)
    line_indices = np.empty(n_lines, dtype=np.int64)
    embedded_in = np.empty(n_lines, dtype=np.int64)
    i = 0
    for info in features:
        point_tags = tags[info.index : info.index + info.size]
        first = point_tags[0]
        for second in point_tags[1:]:
            line_index = gmsh.model.geo.addLine(first, second)
            line_indices[i] = line_index
            embedded_in[i] = info.embedded_in
            first = second
            i += 1
    return line_indices, embedded_in


def add_curve_loop(point_tags: FloatArray) -> int:
    tags = []
    first = point_tags[-1]
    for second in point_tags:
        line_tag = gmsh.model.geo.addLine(first, second)
        tags.append(line_tag)
        first = second
    curve_loop_tag = gmsh.model.geo.addCurveLoop(tags)
    return curve_loop_tag


def add_polygons(
    features: List[PolygonInfo], tags: IntArray
) -> Tuple[List[int], List[int]]:
    plane_tags = []
    for info in features:
        # Add the exterior loop first
        curve_loop_tags = [add_curve_loop(tags[info.index : info.index + info.size])]
        # Now add holes
        for start, size in zip(info.interior_indices, info.interior_sizes):
            loop_tag = add_curve_loop(tags[start : start + size])
            curve_loop_tags.append(loop_tag)
        plane_tag = gmsh.model.geo.addPlaneSurface(curve_loop_tags, tag=info.polygon_id)
        plane_tags.append(plane_tag)
    return curve_loop_tags, plane_tags


def add_points(points: gpd.GeoDataFrame) -> Tuple[IntArray, IntArray]:
    n_points = len(points)
    indices = np.empty(n_points, dtype=np.int64)
    embedded_in = points["__polygon_id"].to_numpy()
    # We have to add points one by one due to the Gmsh addPoint API
    for i, row in enumerate(points.to_dict("records")):
        point = row["geometry"]
        # Rely on the automatic number of gmsh now to generate the indices
        point_index = gmsh.model.geo.addPoint(
            point.x, point.y, Z_DEFAULT, row["cellsize"]
        )
        indices[i] = point_index
    return indices, embedded_in


def collect_polygons(
    polygons: gpd.GeoDataFrame, index: int
) -> Tuple[int, FloatArray, IntArray, List[PolygonInfo]]:
    vertices = []
    cellsizes = []
    features = []
    for row in polygons.to_dict("records"):
        info, coords, cells, index = polygon_info(
            row["geometry"], row["cellsize"], index, row["__polygon_id"]
        )
        vertices.extend(coords)
        cellsizes.extend(cells)
        features.append(info)
    return index, vertices, cellsizes, features


def collect_linestrings(
    linestrings: gpd.GeoDataFrame, index: int
) -> Tuple[int, FloatArray, IntArray, List[LineStringInfo]]:
    vertices = []
    cellsizes = []
    features = []
    for row in linestrings.to_dict("records"):
        info, coords, cells, index = linestring_info(
            row["geometry"], row["cellsize"], index, row["__polygon_id"]
        )
        vertices.append(coords)
        cellsizes.append(cells)
        features.append(info)
    return index, vertices, cellsizes, features


def collect_points(points: gpd.GeoDataFrame) -> FloatArray:
    return np.stack((points["geometry"].x, points["geometry"].y), axis=1)


def embed_where(gdf: gpd.GeoDataFrame, polygons: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    tmp = gpd.sjoin(gdf, polygons, predicate="within", how="inner")
    tmp["cellsize"] = tmp[["cellsize_left", "cellsize_right"]].min(axis=1)
    return tmp[["cellsize", "__polygon_id", "geometry"]]


def add_geometry(
    polygons: gpd.GeoDataFrame, linestrings: gpd.GeoDataFrame, points: gpd.GeoDataFrame
) -> None:
    # Assign unique ids
    polygons["__polygon_id"] = np.arange(1, len(polygons) + 1)

    # Figure out in which polygon the points and linestrings will be embedded.
    linestrings = embed_where(linestrings, polygons)
    embedded_points = embed_where(points, polygons)

    # Collect all coordinates, and store the length and type of every element
    index, poly_vertices, poly_cellsizes, polygon_features = collect_polygons(
        polygons, index=0
    )
    index, line_vertices, line_cellsizes, linestring_features = collect_linestrings(
        linestrings, index
    )
    vertices = np.concatenate(poly_vertices + line_vertices)
    cellsizes = np.concatenate(poly_cellsizes + line_cellsizes)

    # Get the unique vertices, and generate the array of indices pointing to
    # them for every feature
    vertices, indices = np.unique(
        vertices.reshape(-1).view(coord_dtype), return_inverse=True
    )
    vertex_tags = np.arange(1, len(vertices) + 1)
    tags = vertex_tags[indices]
    # Get the smallest cellsize per vertex
    cellsizes = pd.Series(cellsizes).groupby(tags).min().to_numpy()

    # Add all unique vertices. This includes vertices for linestrings and polygons.
    add_vertices(vertices, cellsizes, vertex_tags)
    # Add all geometries to gmsh
    add_polygons(polygon_features, tags)
    linestring_indices, linestring_embedded = add_linestrings(linestring_features, tags)
    gmsh.model.geo.synchronize()

    # Now embed the points and linestrings in the polygons
    for polygon_id, embed_indices in pd.Series(linestring_indices).groupby(
        linestring_embedded
    ):
        gmsh.model.mesh.embed(LINE_DIM, embed_indices, PLANE_DIM, polygon_id)

    if len(embedded_points) > 0:
        point_indices, point_embedded = add_points(embedded_points)
        gmsh.model.geo.synchronize()
        for polygon_id, embed_indices in pd.Series(point_indices).groupby(
            point_embedded
        ):
            gmsh.model.mesh.embed(POINT_DIM, embed_indices, PLANE_DIM, polygon_id)

    gmsh.model.geo.synchronize()
    return


def add_distance_points(points: gpd.GeoSeries) -> IntArray:
    indices = np.empty(len(points), dtype=np.int64)
    for i, (x, y) in enumerate(shapely.get_coordinates(points)):
        indices[i] = gmsh.model.geo.addPoint(x, y, Z_DEFAULT)
    return indices


def add_distance_linestring(
    linestring: sg.LineString,
    distance: float,
) -> IntArray:
    # We could add the line as a Gmsh curve, but Gmsh samples along the line
    # anyway, and the number of points is configured through a discrete
    # sampling value, rather than some spacing distance. So we might as well
    # sample ourselves.
    length = linestring.length
    n = int(np.ceil(length / distance))
    interpolated = shapely.get_coordinates(
        shapely.line_interpolate_point(linestring, distance=np.linspace(0.0, length, n))
    )
    indices = np.empty(len(interpolated), dtype=np.int64)
    for i, (x, y) in enumerate(interpolated):
        indices[i] = gmsh.model.geo.addPoint(x, y, Z_DEFAULT)
    return indices


def add_distance_linestrings(
    linestrings: gpd.GeoSeries,
    spacing: FloatArray,
) -> IntArray:
    indices = [
        add_distance_linestring(linestring, distance)
        for linestring, distance in zip(linestrings, spacing)
    ]
    return np.concatenate(indices)


def add_distance_polygons(polygons: gpd.GeoSeries, spacing: FloatArray) -> IntArray:
    indices = []
    for polygon, distance in zip(polygons, spacing):
        indices.append(add_distance_linestring(polygon.exterior, distance))
        for interior in polygon.interiors:
            indices.append(add_distance_linestring(interior, distance))
    return np.concatenate(indices)


def add_distance_geometry(geometry: gpd.GeoSeries, spacing: FloatArray) -> IntArray:
    polygons, lines, points = separate_geometry(geometry)
    indices = []
    if len(points) > 0:
        indices.append(add_distance_points(points))
    if len(lines) > 0:
        indices.append(add_distance_linestrings(lines, spacing))
    if len(polygons) > 0:
        indices.append(add_distance_polygons(polygons, spacing))
    gmsh.model.geo.synchronize()
    return np.concatenate(indices)



from enum import Enum, EnumMeta
from typing import Any, Type, TypeVar, Union, cast

E = TypeVar("E", bound="FlexibleEnum")


def _show_options(options: Type[E]) -> str:
    return "\n * ".join(map(str, options.__members__))


class AttributeErrorMeta(EnumMeta):
    def __getattr__(cls, name: str) -> Any:
        try:
            return cls.__members__[name]
        except KeyError:
            raise AttributeError(
                f"{name} is not a valid {cls.__name__}. "
                f"Valid options are:\n * {_show_options(cls)}"
            )


class FlexibleEnum(Enum, metaclass=AttributeErrorMeta):
    @classmethod
    def parse(cls: Type[E], value: str) -> E:
        try:
            return cls.__members__[value]
        except KeyError:
            raise ValueError(
                # Use __repr__() so strings are shown with quotes.
                f"{value.__repr__()} is not a valid {cls.__name__}. "
                f"Valid options are:\n * {_show_options(cls)}"
            )

    @classmethod
    def from_value(cls: Type[E], value: Union[E, str]) -> E:
        if isinstance(value, cls):
            return value
        else:
            value = cast(str, value)
            return cls.parse(value)



from pandamesh.enum_base import FlexibleEnum


class DelaunayAlgorithm(FlexibleEnum):
    """The type of Delaunay algorithm for Triangle."""

    DIVIDE_AND_CONQUER = ""
    """Default algorithm."""
    INCREMENTAL = "i"
    """
    Uses the incremental algorithm for Delaunay triangulation, rather than
    the divide-and-conquer algorithm.
    """
    SWEEPLINE = "F"
    """
    Uses Steven Fortune's sweepline algorithm for Delaunay triangulation,
    rather than the divide-and-conquer algorithm.
    """



import pathlib
import tempfile
from contextlib import contextmanager
from typing import Any, List, Tuple, Union

import geopandas as gpd
import numpy as np

from pandamesh.common import (
    FloatArray,
    IntArray,
    central_origin,
    check_geodataframe,
    gmsh,
    move_origin,
    repr,
    separate_geodataframe,
    to_geodataframe,
    to_ugrid,
)
from pandamesh.gmsh_enums import (
    FieldCombination,
    GeneralVerbosity,
    MeshAlgorithm,
    SubdivisionAlgorithm,
)
from pandamesh.gmsh_fields import (
    CombinationField,
    DistanceField,
    GmshField,
    MathEvalField,
    StructuredField,
    ThresholdField,
)
from pandamesh.gmsh_geometry import add_distance_geometry, add_geometry


@contextmanager
def gmsh_env(read_config_files: bool = True, interruptible: bool = True):
    if gmsh.is_initialized() == 1:
        gmsh.finalize()
    try:
        gmsh.initialize(
            readConfigFiles=read_config_files, run=False, interruptible=interruptible
        )
        # There's gotta be a better way to actually raise proper errors...
        gmsh.option.setNumber("General.Terminal", 1)
        yield
    finally:
        gmsh.finalize()


class GmshMesher:
    """
    Wrapper for the python bindings to Gmsh. This class must be initialized
    with a geopandas GeoDataFrame containing at least one polygon, and a column
    named ``"cellsize"``.

    Optionally, multiple polygons with different cell sizes can be included in
    the geodataframe. These can be used to achieve local mesh refinement.

    Linestrings and points may also be included. The segments of linestrings
    will be directly forced into the triangulation. Points can also be forced
    into the triangulation. Unlike Triangle, the cell size values associated
    with these geometries **will** be used.

    Gmsh cannot automatically resolve overlapping polygons, or points
    located exactly on segments. During initialization, the geometries of
    the geodataframe are checked:

        * Polygons should not have any overlap with each other.
        * Linestrings should not intersect each other, unless the intersection
          vertex is present in both.
        * Every linestring should be fully contained by a single polygon;
          a linestring may not intersect two or more polygons.
        * Linestrings and points should not "touch" / be located on
          polygon borders.
        * Holes in polygons are fully supported, but they must not contain
          any linestrings or points.

    If such cases are detected, the initialization will error: use the
    :class:`pandamesh.Preprocessor` to clean up geometries beforehand.

    For more details on Gmsh, see:
    https://gmsh.info/doc/texinfo/gmsh.html

    A helpful index can be found near the bottom:
    https://gmsh.info/doc/texinfo/gmsh.html#Syntax-index

    .. note::

        This meshers uses the Gmsh Python API, which is global. To avoid a
        situation where multiple GmshMeshers have been iniatilized and are
        mutating each other's (global) variables, the ``.finalize()`` method
        must be called before instantiating a new mesher.

    Parameters
    ----------
    gdf: gpd.GeoDataFrame
        GeoDataFrame containing the vector geometry. Must contain a "cellsize"
        column.
    shift_origin: bool, optional, default is True.
        If True, temporarily shifts the coordinate system origin to the centroid
        of the geometry's bounding box during mesh generation. This helps mitigate
        floating-point precision issues. The resulting mesh vertices are
        automatically translated back to the original coordinate system.
    intersecting_edges: str, optional, default is "error"
        String indicating how to report unresolved line segment intersections:

        * "ignore": skip check.
        * "warning": emit a warning.
        * "error": raise a ValueError.
    read_config_files: bool
        Gmsh initialization option: Read system Gmsh configuration files
        (gmshrc and gmsh-options).
    interruptible: bool
        Gmsh initialization option.
    """

    _instance = None

    def __new__(cls, *args, **kwargs) -> "GmshMesher":
        # The Gmsh Python API is unfortunately global. Make sure only one
        # GmshMesher instance is active at a given time.
        if cls._instance is not None:
            raise RuntimeError(
                "Multiple GmshMesher instances are not allowed. "
                "Please use either:\n"
                "- existing_instance.finalize()\n"
                "- GmshMesher.finalize()\n"
                "to clean up before creating a new instance."
            )
        cls._instance = super().__new__(cls)
        cls._instance._initialized = True
        return cls._instance

    @classmethod
    def finalize(cls) -> None:
        # Finalize gmsh, set flag to uninitialized
        try:
            cls.finalize_gmsh()
        finally:
            if cls._instance is not None:
                cls._instance._initialized = False
                cls._instance = None

    def __getattribute__(self, name: str) -> Any:
        # Make sure to error if finalized has already been called.
        if name in ("__dict__", "_initialized", "_initialize_gmsh", "finalize"):
            return super().__getattribute__(name)
        if not self._initialized:
            raise RuntimeError("This GmshMesher has been finalized")
        return super().__getattribute__(name)

    @classmethod
    def _force_init(
        cls,
        gdf: gpd.GeoDataFrame,
        shift_origin: bool = True,
        intersecting_edges: str = "error",
        read_config_files: bool = True,
        interruptible: bool = True,
    ) -> "GmshMesher":
        """
        Guarantees a new instance.

        This method ensures that only one instance of GmshMesher exists at a
        time. If an instance already exists, it is finalized before this method
        creates a new one.

        Should be used only for testing.
        """
        cls.finalize()
        return cls(
            gdf, shift_origin, intersecting_edges, read_config_files, interruptible
        )

    def __init__(
        self,
        gdf: gpd.GeoDataFrame,
        shift_origin: bool = True,
        intersecting_edges="error",
        read_config_files: bool = True,
        interruptible: bool = True,
    ) -> None:
        self._initialize_gmsh(
            read_config_files=read_config_files, interruptible=interruptible
        )
        check_geodataframe(gdf, {"geometry", "cellsize"}, check_index=True)
        gdf, self._xoff, self._yoff = central_origin(gdf, shift_origin)
        polygons, linestrings, points = separate_geodataframe(gdf, intersecting_edges)

        # Include geometry into gmsh
        add_geometry(polygons, linestrings, points)

        # Initialize fields parameters
        self._fields: List[GmshField] = []
        self._combination_field = None
        self._tmpdir = tempfile.TemporaryDirectory()

        # Set default values for meshing parameters
        self.mesh_algorithm = MeshAlgorithm.AUTOMATIC
        self.recombine_all = False
        self.mesh_size_extend_from_boundary = True
        self.mesh_size_from_points = True
        self.mesh_size_from_curvature = False
        self.field_combination = FieldCombination.MIN
        self.subdivision_algorithm = SubdivisionAlgorithm.NONE
        self.general_verbosity = GeneralVerbosity.SILENT

    def __repr__(self):
        return repr(self)

    @staticmethod
    def _initialize_gmsh(read_config_files: bool = True, interruptible: bool = True):
        GmshMesher.finalize_gmsh()
        gmsh.initialize(
            readConfigFiles=read_config_files, run=False, interruptible=interruptible
        )
        gmsh.option.setNumber("General.Terminal", 1)

    @staticmethod
    def finalize_gmsh():
        """Finalize Gmsh."""
        if gmsh.is_initialized() == 1:
            gmsh.finalize()

    # Properties
    # ----------
    @property
    def mesh_algorithm(self) -> MeshAlgorithm:
        """
        Can be set to one of :py:class:`pandamesh.MeshAlgorithm`:

        .. code::

            MESH_ADAPT = 1
            AUTOMATIC = 2
            INITIAL_MESH_ONLY = 3
            FRONTAL_DELAUNAY = 5
            BAMG = 7
            FRONTAL_DELAUNAY_FOR_QUADS = 8
            PACKING_OF_PARALLELLOGRAMS = 9
            QUASI_STRUCTURED_QUAD = 11

        Each algorithm has its own advantages and disadvantages.
        """
        return MeshAlgorithm(gmsh.option.getNumber("Mesh.Algorithm"))

    @mesh_algorithm.setter
    def mesh_algorithm(self, value: MeshAlgorithm):
        value = MeshAlgorithm.from_value(value)
        gmsh.option.setNumber("Mesh.Algorithm", value.value)

    @property
    def recombine_all(self) -> bool:
        """
        Apply recombination algorithm to all surfaces, ignoring per-surface
        spec.
        """
        return self._recombine_all

    @recombine_all.setter
    def recombine_all(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise TypeError("recombine_all must be a bool")
        self._recombine_all = value
        gmsh.option.setNumber("Mesh.RecombineAll", value)

    @property
    def mesh_size_extend_from_boundary(self) -> bool:
        """
        Forces the mesh size to be extended from the boundary, or not, per
        surface.
        """
        return self._mesh_size_extend_from_boundary

    @mesh_size_extend_from_boundary.setter
    def mesh_size_extend_from_boundary(self, value):
        if not isinstance(value, bool):
            raise TypeError("mesh_size_extend_from_boundary must be a bool")
        self._mesh_size_extend_from_boundary = value
        gmsh.option.setNumber("Mesh.MeshSizeExtendFromBoundary", value)

    @property
    def mesh_size_from_points(self):
        """Compute mesh element sizes from values given at geometry points."""
        return self._mesh_size_from_points

    @mesh_size_from_points.setter
    def mesh_size_from_points(self, value):
        if not isinstance(value, bool):
            raise TypeError("mesh_size_from_points must be a bool")
        self._mesh_size_from_points = value
        gmsh.option.setNumber("Mesh.MeshSizeFromPoints", value)

    @property
    def mesh_size_from_curvature(self):
        """
        Automatically compute mesh element sizes from curvature, using the value as
        the target number of elements per 2 * Pi radians.
        """
        return self._mesh_size_from_curvature

    @mesh_size_from_curvature.setter
    def mesh_size_from_curvature(self, value):
        """
        Automatically compute mesh element sizes from curvature, using the
        value as the target number of elements per 2 * Pi radians.
        """
        if not isinstance(value, bool):
            raise TypeError("mesh_size_from_curvature must be a bool")
        self._mesh_size_from_curvature = value
        gmsh.option.setNumber("Mesh.MeshSizeFromCurvature", value)

    @property
    def field_combination(self) -> FieldCombination:
        """
        Controls how cell size fields are combined when they are found at the
        same location. Can be set to one of
        :py:class:`pandamesh.FieldCombination`:

        .. code::

            MIN = "Min"
            MAX = "Max"

        """
        return self._field_combination

    @field_combination.setter
    def field_combination(self, value: Union[FieldCombination, str]):
        value = FieldCombination.from_value(value)
        self._field_combination = value
        # Value is propagated to gmsh in ._combine_fields()

    @property
    def subdivision_algorithm(self) -> SubdivisionAlgorithm:
        """
        All meshes can be subdivided to generate fully quadrangular cells. Can
        be set to one of :py:class:`pandamesh.SubdivisionAlgorithm`:

        .. code::

            NONE = 0
            ALL_QUADRANGLES = 1
            BARYCENTRIC = 3

        """

        return SubdivisionAlgorithm(gmsh.option.getNumber("Mesh.SubdivisionAlgorithm"))

    @subdivision_algorithm.setter
    def subdivision_algorithm(self, value: Union[SubdivisionAlgorithm, str]):
        value = SubdivisionAlgorithm.from_value(value)
        gmsh.option.setNumber("Mesh.SubdivisionAlgorithm", value.value)

    @property
    def general_verbosity(self) -> GeneralVerbosity:
        """
        Controls level of information printed. Can be set to one of
        :py:class:`pandamesh.GeneralVerbosity`:

        .. code::

            SILENT = 0
            ERRORS = 1
            WARNINGS = 2
            DIRECT = 3
            INFORMATION = 4
            STATUS = 5
            DEBUG = 99

        """
        return GeneralVerbosity(gmsh.option.getNumber("General.Verbosity"))

    @general_verbosity.setter
    def general_verbosity(self, value: Union[GeneralVerbosity, str]) -> None:
        value = GeneralVerbosity.from_value(value)
        gmsh.option.setNumber("General.Verbosity", value.value)

    @property
    def fields(self):
        """
        Read-only access to fields.

        Use ``.clear_fields`` to remove fields from the mesher.
        """
        return self._fields.copy()

    # Methods
    # -------

    def clear_fields(self):
        while self._fields:
            field = self._fields.pop()
            field.remove_from_gmsh()
        if self._combination_field is not None:
            self._combination_field.remove_from_gmsh()
            self._combination_field = None

    def add_matheval_distance_field(self, gdf: gpd.GeoDataFrame) -> None:
        """
        Add a matheval distance field to the mesher.

        The of geometry of these fields are not forced into the mesh, but they
        are used to specify zones of with cell sizes.

        Uses the MathEval functionality in Gmsh, which relies on the SSCILIB
        math expression evaluator.

        https://gitlab.onelab.info/gmsh/gmsh/-/blob/master/contrib/MathEx/mathex.cpp

        https://sscilib.sourceforge.net/

        Parameters
        ----------
        gdf: geopandas.GeoDataFrame
            Location of the features to measure distance to. Should contain
            ``spacing`` column to specify the spacing of interpolated vertices
            along linestrings and polygon boundaries, and a ``function`` column
            to specify the function to control cell size as a function of
            distance. See the examples.

        Examples
        --------
        Generate a number of points:

        >>> x = np.arange(0.0, 10.0)
        >>> y = np.arange(0.0, 10.0)
        >>> points = gpd.points_from_xy(x, y)
        >>> field = gpd.GeoDataFrame(geometry=points)

        Add spacing (dummy value for points) and a function:

        >>> field["spacing"] = np.nan
        >>> field["function"] = "max(distance^2, 1.0)"

        Apply it:

        >>> mesher.add_matheval_distance_field(field)

        The following mathematical operators are supported:

        Basic Operators

        - Arithmetic: ``+``, ``-``, ``*``, ``/``, ``%`` (modulo), ``^`` (power)
        - Comparison: ````<``, ``>``

        Mathematical Functions

        - Absolute value: ``abs(x)``
        - Square root: ``sqrt(x)``
        - Exponential: ``exp(x)``
        - Natural logarithm: ``log(x)``
        - Base-10 logarithm: ``log10(x)``
        - Power: ``pow(x,y)``

        Statistical Functions

        - Minimum: ``min(x, y, ...)``
        - Maximum: ``max(x, y, ...)``
        - Sum: ``sum(x, y, ...)``
        - Average: ``med(x, y, ...)```

        Trigonometric Functions

        - Standard: ``sin(x)``, ``cos(x)``, ``tan(x)``
        - Inverse: ``asin(x)``, ``acos(x)``, ``atan(x)``
        - Hyperbolic: ``sinh(x)``, ``cosh(x)``, ``tanh(x)``

        Rounding Functions

        - ``floor(x)``, ``ceil(x)``, ``round(x)``, ``trunc(x)``

        Constants

        - Pi: ``pi``
        - Euler's number: ``e``
        """
        check_geodataframe(gdf, {"geometry", "function", "spacing"})
        gdf = move_origin(gdf, xoff=self._xoff, yoff=self._yoff)

        for function, group in gdf.groupby("function"):
            point_tags = add_distance_geometry(group["geometry"], group["spacing"])
            distance_field = DistanceField(point_tags)
            math_eval_field = MathEvalField(distance_field, function)
            self._fields.append(math_eval_field)
        return

    def add_threshold_distance_field(self, gdf: gpd.GeoDataFrame) -> None:
        """
        Add a distance field to the mesher.

        The of geometry of these fields are not forced into the mesh, but they
        can be used to specify zones of with cell sizes.

        Parameters
        ----------
        gdf: geopandas.GeoDataFrame
            Location of the features to measure distance to. Should contain
            ``spacing`` column to specify the spacing of interpolated vertices
            along linestrings and polygon boundaries.
        """
        columns = ["size_min", "size_max", "dist_min", "dist_max"]
        check_geodataframe(gdf, set(columns + ["geometry", "spacing"]))
        gdf = move_origin(gdf, xoff=self._xoff, yoff=self._yoff)

        if "sigmoid" not in gdf.columns:
            gdf["sigmoid"] = False
        if "stop_at_dist_max" not in gdf.columns:
            gdf["stop_at_dist_max"] = False
        columns += ["sigmoid", "stop_at_dist_max"]

        grouped = gdf.groupby(columns)
        for (
            size_min,
            size_max,
            dist_min,
            dist_max,
            sigmoid,
            stop_at_dist_max,
        ), group in grouped:
            point_tags = add_distance_geometry(group["geometry"], group["spacing"])
            distance_field = DistanceField(point_tags)
            threshold_field = ThresholdField(
                distance_field=distance_field,
                size_min=size_min,
                size_max=size_max,
                dist_min=dist_min,
                dist_max=dist_max,
                sigmoid=sigmoid,
                stop_at_dist_max=stop_at_dist_max,
            )
            self._fields.append(threshold_field)

        return

    def add_structured_field(
        self,
        cellsize: FloatArray,
        xmin: float,
        ymin: float,
        dx: float,
        dy: float,
        outside_value: Union[float, None] = None,
    ) -> None:
        """
        Add an equidistant structured field specifying cell sizes. Gmsh will
        interpolate between the points to determine the desired cell size.

        Parameters
        ----------
        cellsize: FloatArray with shape ``(n_y, n_x``)
            Specifies the cell size on a structured grid. The location of this grid
            is determined by ``xmin, ymin, dx, dy``.
        xmin: float
            x-origin.
        ymin: float
            y-origin.
        dx: float
            Spacing along the x-axis.
        dy: float
            Spacing along the y-axis.
        outside_value: Union[float, None]
            Value outside of the window ``(xmin, xmax)`` and ``(ymin, ymax)``.
            Default value is None.
        """
        self._fields.append(
            StructuredField(
                tmpdir=self._tmpdir,
                cellsize=cellsize,
                xmin=xmin - self._xoff,
                ymin=ymin - self._yoff,
                dx=dx,
                dy=dy,
                outside_value=outside_value,
            )
        )

    def add_structured_field_from_dataarray(
        self,
        da: "xarray.DataArray",  # noqa: F821
        outside_value: Union[float, None] = None,
    ):
        """
        Add an equidistant structured field specifying cell sizes from an
        xarray DataArray object. Gmsh will interpolate between the points to
        determine the desired cell size.

        Parameters
        ----------
        da: xarray.DataArray
            Values are used as cell sizes. Must have dimensions `("y", "x")`.
        outside_value: Union[float, None]
            Value outside of the window ``(xmin, xmax)`` and ``(ymin, ymax)``.
            Default value is None.
        """
        import xarray as xr

        if not isinstance(da, xr.DataArray):
            raise TypeError(
                f"da must be xr.DataArray, received instead: {type(da).__name__}"
            )
        if not da.dims == ("y", "x"):
            raise ValueError(f'Dimensions must be ("y", "x"), received: {da.dims}')

        dxs = np.diff(da["x"]) if len(da["x"]) > 0 else np.array([0.0])
        dys = np.diff(da["y"]) if len(da["y"]) > 0 else np.array([0.0])
        dx = dxs[0]
        dy = dys[0]

        if not np.allclose(dxs, dx, atol=1e-4 * dx):
            raise ValueError("da is not equidistant along x")
        if not np.allclose(dys, dx, atol=1e-4 * dy):
            raise ValueError("da is not equidistant along y")

        self.add_structured_field(
            cellsize=da.to_numpy(),
            xmin=float(da["x"].min()),
            ymin=float(da["y"].min()),
            dx=dx,
            dy=dy,
            outside_value=outside_value,
        )

    def _combine_fields(self) -> None:
        if self._combination_field is not None:
            self._combination_field.remove_from_gmsh()
        self._combination_field = CombinationField(self._fields, self.field_combination)

    def _vertices(self):
        # getNodes returns: node_tags, coord, parametric_coord
        _, vertices, _ = gmsh.model.mesh.getNodes()
        # Return x and y
        vertices = vertices.reshape((-1, 3))[:, :2]
        vertices[:, 0] += self._xoff
        vertices[:, 1] += self._yoff
        return np.ascontiguousarray(vertices)

    def _faces(self):
        element_types, _, node_tags = gmsh.model.mesh.getElements()
        tags = dict(zip(element_types, node_tags))
        _TRIANGLE = 2
        _QUAD = 3
        _FILL_VALUE = 0
        # Combine triangle and quad faces if the mesh is heterogenous
        if _TRIANGLE in tags and _QUAD in tags:
            triangle_faces = tags[_TRIANGLE].reshape((-1, 3))
            quad_faces = tags[_QUAD].reshape((-1, 4))
            n_triangle = triangle_faces.shape[0]
            n_quad = quad_faces.shape[0]
            faces = np.full((n_triangle + n_quad, 4), _FILL_VALUE)
            faces[:n_triangle, :3] = triangle_faces
            faces[n_triangle:, :] = quad_faces
        elif _QUAD in tags:
            faces = tags[_QUAD].reshape((-1, 4))
        elif _TRIANGLE in tags:
            faces = tags[_TRIANGLE].reshape((-1, 3))
        else:
            raise ValueError("No triangles or quads in mesh")
        # convert to 0-based index
        return faces - 1

    def generate(self, finalize: bool = False) -> Tuple[FloatArray, IntArray]:
        """
        Generate a mesh of triangles or quadrangles.

        Parameters
        ----------
        finalize: bool, default False
            Automatically finalize after generating.

        Returns
        -------
        vertices: np.ndarray of floats with shape ``(n_vertex, 2)``
        faces: np.ndarray of integers with shape ``(n_face, nmax_per_face)``
            ``nmax_per_face`` is 3 for exclusively triangles and 4 if
            quadrangles are included. A fill value of -1 is used as a last
            entry for triangles in that case.
        """
        try:
            # Remove any previously generated results from gmsh.
            gmsh.model.mesh.clear()
            self._combine_fields()
            gmsh.model.mesh.generate(dim=2)

            # cleaning up of mesh in order to obtain unique elements and nodes
            gmsh.model.mesh.removeDuplicateElements()
            gmsh.model.mesh.removeDuplicateNodes()
            gmsh.model.mesh.renumberElements()
            gmsh.model.mesh.renumberNodes()

            vertices, faces = self._vertices(), self._faces()
        finally:
            if finalize:
                self.finalize()
        return vertices, faces

    def generate_geodataframe(self, finalize: bool = False) -> gpd.GeoDataFrame:
        """
        Generate a mesh and return it as a geopandas GeoDataFrame.

        Parameters
        ----------
        finalize: bool, default False
            Automatically finalize after generating.

        Returns
        -------
        mesh: geopandas.GeoDataFrame
        """
        return to_geodataframe(*self.generate(finalize=finalize))

    def generate_ugrid(self, finalize: bool = False) -> "xugrid.Ugrid2d":  # type: ignore # noqa  pragma: no cover
        """
        Generate a mesh and return it as an xugrid Ugrid2d.

        Parameters
        ----------
        finalize: bool, default False
            Automatically finalize after generating.

        Returns
        -------
        mesh: xugrid.Ugrid2d
        """
        return to_ugrid(*self.generate(finalize=finalize))

    def write(self, path: Union[str, pathlib.Path]):
        """
        Write a gmsh .msh file

        Parameters
        ----------
        path: Union[str, pathlib.Path
        """
        gmsh.write(str(path))



import os

import numpy as np
import pytest
import shapely

from pandamesh.snapping import MatrixCSR, columns_and_values, row_slice, snap_nodes


def numba_enabled() -> bool:
    return os.environ.get("NUMBA_DISABLE_JIT") != "1"


@pytest.fixture(scope="function")
def csr_matrix():
    i = np.repeat(np.arange(5), 2)
    j = np.arange(10)
    v = np.full(10, 0.5)
    return MatrixCSR.from_triplet(i, j, v, 5, 10)


@pytest.mark.skipif(
    numba_enabled(),
    reason="Function returns a slice object; python and no-python slices don't mix.",
)
def test_row_slice(csr_matrix):
    # These functions work fine if called inside of other numba functions when
    # numba is enabled.
    assert row_slice(csr_matrix, 0) == slice(0, 2, None)


@pytest.mark.skipif(
    numba_enabled(),
    reason="Function returns a zip object; python and no-python zips don't mix.",
)
def test_columns_and_values(csr_matrix):
    # These functions work fine if called inside of other numba functions when
    # numba is enabled.
    zipped = columns_and_values(csr_matrix, row_slice(csr_matrix, 0))
    result = list(zipped)
    assert result == [(0, 0.5), (1, 0.5)]


def test_snap__three_points_horizontal():
    x = np.array([0.0, 1.0, 2.0])
    y = np.zeros_like(x)
    xy = shapely.points(x, y)
    index = snap_nodes(xy, 0.1)
    assert np.array_equal(index, [0, 1, 2])

    index = snap_nodes(xy, 1.0)
    assert np.array_equal(index, [0, 2])

    index = snap_nodes(xy, 2.0)
    assert np.array_equal(index, [0])


def test_snap__three_points_diagonal():
    x = y = np.array([0.0, 1.0, 1.5])
    xy = shapely.points(x, y)
    index = snap_nodes(xy, 0.1)
    assert np.array_equal(index, [0, 1, 2])

    # hypot(0.5, 0.5) = 0.707...
    index = snap_nodes(xy, 0.71)
    assert np.array_equal(index, [0, 1])

    # hypot(1, 1) = 1.414...
    index = snap_nodes(xy, 1.42)
    assert np.array_equal(index, [0, 2])



import geopandas as gpd
import numpy as np
import shapely
import shapely.geometry as sg

from pandamesh import triangle_geometry as tg

outer_coords = np.array([(0.0, 0.0), (10.0, 0.0), (10.0, 10.0), (0.0, 10.0)])
inner_coords = np.array([(3.0, 3.0), (7.0, 3.0), (7.0, 7.0), (3.0, 7.0)])
ring_coords = np.array([(3.0, 3.0), (7.0, 3.0), (7.0, 7.0), (3.0, 7.0), (3.0, 3.0)])
line_coords = np.array([(2.0, 8.0), (8.0, 2.0)])
inner = sg.LinearRing(inner_coords)
outer = sg.LinearRing(outer_coords)
line = sg.LineString(line_coords)
ring = sg.LineString(ring_coords)
donut = sg.Polygon(outer, holes=[inner])
refined = sg.Polygon(inner_coords)


def area(vertices, triangles):
    """
    Compute the area of every triangle in the mesh.
    (Helper for these tests.)
    """
    coords = vertices[triangles]
    u = coords[:, 1] - coords[:, 0]
    v = coords[:, 2] - coords[:, 0]
    return 0.5 * np.abs(u[:, 0] * v[:, 1] - u[:, 1] * v[:, 0])


def test_add_linestrings():
    series = gpd.GeoSeries(data=[line])
    vertices, segments = tg.add_linestrings(series)
    expected = np.unique(line_coords, axis=0)
    expected_segments = np.array([[0, 1]])
    assert np.allclose(vertices, expected)
    assert np.array_equal(segments, expected_segments)

    series = gpd.GeoSeries(data=[inner])
    vertices, segments = tg.add_linestrings(series)
    vertices, segments = tg.unique_vertices_and_segments(vertices, segments)
    expected = np.unique(inner_coords, axis=0)
    expected_segments = np.array(
        [
            [0, 2],
            [1, 0],
            [2, 3],
            [3, 1],
        ]
    )
    assert np.allclose(vertices, expected)
    assert np.array_equal(segments, expected_segments)

    series = gpd.GeoSeries(data=[outer])
    vertices, segments = tg.add_linestrings(series)
    vertices, segments = tg.unique_vertices_and_segments(vertices, segments)
    expected = np.unique(outer_coords, axis=0)
    assert np.allclose(vertices, expected)
    assert np.array_equal(segments, expected_segments)

    # Empty should work too
    series = gpd.GeoSeries(data=[])
    _, _ = tg.add_linestrings(series)


def test_add_polygons():
    gdf = gpd.GeoDataFrame(geometry=[donut])
    cellsize = 0.5
    gdf["cellsize"] = cellsize
    vertices, segments, regions = tg.add_polygons(gdf)
    vertices, segments = tg.unique_vertices_and_segments(vertices, segments)
    expected = np.unique(
        np.concatenate([outer_coords, inner_coords]),
        axis=0,
    )
    expected_segments = np.array(
        [
            [0, 6],
            [1, 0],
            [2, 4],
            [3, 2],
            [4, 5],
            [5, 3],
            [6, 7],
            [7, 1],
        ]
    )
    x, y = regions[0, :2]
    assert np.allclose(vertices, expected)
    assert np.array_equal(segments, expected_segments)
    assert regions[0, 2] == 0
    assert np.allclose(regions[0, 3], 0.25 * np.sqrt(3) * cellsize**2)
    assert sg.Point(x, y).within(donut)


def test_add_points():
    xy = np.array(
        [
            [0.0, 0.0],
            [1.0, 1.0],
        ]
    )
    gdf = gpd.GeoDataFrame(geometry=gpd.points_from_xy(xy[:, 0], xy[:, 1]))
    vertices = tg.add_points(gdf)
    assert np.allclose(vertices, xy)


def test_polygon_holes():
    polygon = sg.Polygon(outer)
    gdf = gpd.GeoDataFrame(geometry=[polygon])
    assert tg.polygon_holes(gdf) is None

    gdf = gpd.GeoDataFrame(geometry=[donut])
    assert len(tg.polygon_holes(gdf)) == 1

    gdf = gpd.GeoDataFrame(geometry=[donut, refined])
    assert tg.polygon_holes(gdf) is None


def test_convert_ring_linestring():
    # The linestring forms a ring within the outer polygon. During the
    # conversion, the ring should be converted to a second polygon, and a hole
    # should be made in the first polygon.
    polygon = sg.Polygon(shell=outer_coords)
    polygons = gpd.GeoDataFrame(geometry=[polygon])
    polygons["cellsize"] = 1.0
    linestrings = gpd.GeoDataFrame(geometry=[ring])

    new_polygons = tg.convert_linestring_rings(polygons, linestrings)
    assert isinstance(new_polygons, gpd.GeoDataFrame)
    assert np.allclose(new_polygons["cellsize"], 1.0)
    assert np.allclose(new_polygons.area, [84.0, 16.0])


def test_convert_ring_linestring__hole():
    # This second case has a hole inside of the linestring ring. The hole
    # should be preserved.
    inner = [
        [4.0, 6.0],
        [6.0, 6.0],
        [6.0, 4.0],
        [4.0, 4.0],
        [4.0, 6.0],
    ]
    polygon = sg.Polygon(shell=outer_coords, holes=[inner])
    polygons = gpd.GeoDataFrame(geometry=[polygon])
    polygons["cellsize"] = 1.0
    linestrings = gpd.GeoDataFrame(geometry=[ring])

    new_polygons = tg.convert_linestring_rings(polygons, linestrings)
    assert isinstance(new_polygons, gpd.GeoDataFrame)
    assert np.allclose(new_polygons["cellsize"], 1.0)
    assert np.allclose(new_polygons.area, [84.0, 12.0])


def test_convert_ring_linestring__nested_ring():
    # This third case has the ring inside of a nested polygon.
    nested_ring = [
        [4.0, 6.0],
        [6.0, 6.0],
        [6.0, 4.0],
        [4.0, 4.0],
        [4.0, 6.0],
    ]
    polygon = sg.Polygon(shell=outer_coords, holes=[inner])
    inner_polygon = sg.Polygon(shell=inner_coords)
    polygons = gpd.GeoDataFrame(geometry=[polygon, inner_polygon])
    polygons["cellsize"] = 1.0
    linestrings = gpd.GeoDataFrame(geometry=[sg.LineString(nested_ring)])

    new_polygons = tg.convert_linestring_rings(polygons, linestrings)
    assert isinstance(new_polygons, gpd.GeoDataFrame)
    assert np.allclose(new_polygons["cellsize"], 1.0)
    assert np.allclose(new_polygons.area, [84.0, 12.0, 4.0])


def test_segmentize_linestrings():
    gdf = gpd.GeoDataFrame()
    actual = tg.segmentize_linestrings(gdf)
    assert actual is gdf

    gdf = gpd.GeoDataFrame(
        geometry=[
            sg.LineString([[0.0, 0.0], [10.0, 0.0]]),
            sg.LineString([[0.0, 5.0], [10.0, 5.0]]),
            sg.LineString([[0.0, 10.0], [10.0, 10.0]]),
        ],
        data={"cellsize": [np.nan, 1.0, 0.5]},
    )
    actual = tg.segmentize_linestrings(gdf)
    _, index = shapely.get_coordinates(actual.geometry, return_index=True)
    _, counts = np.unique(index, return_counts=True)
    assert np.array_equal(counts, [2, 11, 21])



import geopandas as gpd
import numpy as np
import pytest
import shapely.geometry as sg

import pandamesh as pm

outer_coords = np.array([(0.0, 0.0), (10.0, 0.0), (10.0, 10.0), (0.0, 10.0)])
inner_coords = np.array([(3.0, 3.0), (7.0, 3.0), (7.0, 7.0), (3.0, 7.0)])
line_coords = np.array([(2.0, 8.0), (8.0, 2.0)])
inner = sg.LinearRing(inner_coords)
outer = sg.LinearRing(outer_coords)
line = sg.LineString(line_coords)
polygon = sg.Polygon(outer)
donut = sg.Polygon(outer, holes=[inner])

other_inner_coords = np.array([(3.0, 4.0), (7.0, 4.0), (7.0, 6.0), (3.0, 6.0)])
other_inner = sg.Polygon(other_inner_coords)
other_hole_coords = np.array(
    [
        (7.0, 3.0),
        (7.0, 4.0),
        (7.0, 6.0),
        (7.0, 7.0),
        (3.0, 7.0),
        (3.0, 6.0),
        (3.0, 4.0),
        (3.0, 3.0),
    ]
)
other_hole = sg.LinearRing(other_hole_coords)
other_donut = sg.Polygon(outer, holes=[other_hole])


def test_singleton_gmsh_mesher():
    pm.GmshMesher.finalize()

    gdf = gpd.GeoDataFrame(geometry=[donut])
    gdf["cellsize"] = 1.0
    mesher = pm.GmshMesher(gdf)

    with pytest.raises(
        RuntimeError, match="Multiple GmshMesher instances are not allowed"
    ):
        pm.GmshMesher(gdf)

    mesher.finalize()
    new_mesher = pm.GmshMesher(gdf)
    assert isinstance(new_mesher, pm.GmshMesher)

    with pytest.raises(RuntimeError, match="This GmshMesher has been finalized"):
        mesher.generate()

    new_mesher.generate(finalize=True)
    assert not new_mesher._initialized


def bounds(vertices):
    x, y = vertices.T
    return x.min(), y.min(), x.max(), y.max()


def area(vertices, triangles):
    """
    Compute the area of every triangle in the mesh.
    (Helper for these tests.)
    """
    coords = vertices[triangles]
    u = coords[:, 1] - coords[:, 0]
    v = coords[:, 2] - coords[:, 0]
    return 0.5 * np.abs(u[:, 0] * v[:, 1] - u[:, 1] * v[:, 0])


def triangle_generate(gdf: gpd.GeoDataFrame, shift: bool):
    mesher = pm.TriangleMesher(gdf, shift_origin=shift)
    return mesher.generate()


def gmsh_generate(gdf: gpd.GeoDataFrame, shift: bool):
    mesher = pm.GmshMesher._force_init(gdf, shift_origin=shift)
    vertices, faces = mesher.generate()
    return vertices, faces


@pytest.mark.parametrize("generate", [triangle_generate, gmsh_generate])
@pytest.mark.parametrize("shift", [False, True])
def test_empty(generate, shift):
    gdf = gpd.GeoDataFrame(geometry=[line], data={"cellsize": [1.0]})
    with pytest.raises(ValueError, match="No polygons provided"):
        generate(gdf, shift)


@pytest.mark.parametrize("generate", [triangle_generate, gmsh_generate])
@pytest.mark.parametrize("shift", [False, True])
def test_basic(generate, shift):
    gdf = gpd.GeoDataFrame(geometry=[polygon])
    gdf["cellsize"] = 1.0
    vertices, triangles = generate(gdf, shift)
    mesh_area = area(vertices, triangles).sum()
    assert np.allclose(mesh_area, polygon.area)
    assert np.allclose(bounds(vertices), gdf.total_bounds)


@pytest.mark.parametrize("generate", [triangle_generate, gmsh_generate])
@pytest.mark.parametrize("shift", [False, True])
def test_hole(generate, shift):
    gdf = gpd.GeoDataFrame(geometry=[donut])
    gdf["cellsize"] = 1.0
    vertices, triangles = generate(gdf, shift)
    mesh_area = area(vertices, triangles).sum()
    assert np.allclose(mesh_area, donut.area)
    assert np.allclose(bounds(vertices), gdf.total_bounds)


@pytest.mark.parametrize("generate", [triangle_generate, gmsh_generate])
@pytest.mark.parametrize("shift", [False, True])
def test_ring(generate, shift):
    gdf = gpd.GeoDataFrame(geometry=[polygon, inner])
    gdf["cellsize"] = 1.0
    vertices, triangles = generate(gdf, shift)
    mesh_area = area(vertices, triangles).sum()
    assert np.allclose(mesh_area, polygon.area)
    assert np.allclose(bounds(vertices), gdf.total_bounds)


@pytest.mark.parametrize("generate", [triangle_generate, gmsh_generate])
@pytest.mark.parametrize("shift", [False, True])
def test_partial_hole(generate, shift):
    gdf = gpd.GeoDataFrame(geometry=[other_donut, other_inner])
    gdf["cellsize"] = 1.0
    vertices, triangles = generate(gdf, shift)
    mesh_area = area(vertices, triangles).sum()
    assert np.allclose(mesh_area, other_donut.area + other_inner.area)
    assert np.allclose(bounds(vertices), gdf.total_bounds)


@pytest.mark.parametrize("generate", [triangle_generate, gmsh_generate])
@pytest.mark.parametrize("shift", [False, True])
def test_adjacent_donut(generate, shift):
    inner_coords2 = inner_coords.copy()
    outer_coords2 = outer_coords.copy()
    inner_coords2[:, 0] += 10.0
    outer_coords2[:, 0] += 10.0
    inner2 = sg.LinearRing(inner_coords2)
    outer2 = sg.LinearRing(outer_coords2)
    donut2 = sg.Polygon(outer2, holes=[inner2])

    gdf = gpd.GeoDataFrame(geometry=[donut, donut2])
    gdf["cellsize"] = [1.0, 0.5]
    vertices, triangles = generate(gdf, shift)
    mesh_area = area(vertices, triangles).sum()
    assert np.allclose(mesh_area, 2 * donut.area)
    assert np.allclose(bounds(vertices), gdf.total_bounds)

    # With a line at y=8.0 and points in the left polygon, at y=2.0
    line1 = sg.LineString([(0.25, 8.0), (9.75, 8.0)])
    line2 = sg.LineString([(10.25, 8.0), (19.75, 8.0)])
    x = np.arange(0.25, 10.0, 0.25)
    y = np.full(x.size, 2.0)
    points = gpd.points_from_xy(x=x, y=y)
    gdf = gpd.GeoDataFrame(geometry=[donut, donut2, line1, line2, *points])
    gdf["cellsize"] = 1.0

    vertices, triangles = generate(gdf, shift)
    mesh_area = area(vertices, triangles).sum()
    assert np.allclose(mesh_area, 2 * donut.area)
    assert np.allclose(bounds(vertices), gdf.total_bounds)


def test_triangle_properties():
    gdf = gpd.GeoDataFrame(geometry=[donut])
    gdf["cellsize"] = 1.0
    mesher = pm.TriangleMesher(gdf)

    # Should be a float >=0, < 34.0
    with pytest.raises(TypeError):
        mesher.minimum_angle = 10
    with pytest.raises(ValueError):
        mesher.minimum_angle = 35.0

    # Set properties
    mesher.minimum_angle = 10.0
    mesher.conforming_delaunay = False
    mesher.suppress_exact_arithmetic = True
    mesher.maximum_steiner_points = 10
    mesher.delaunay_algorithm = pm.DelaunayAlgorithm.SWEEPLINE
    mesher.consistency_check = True

    # Check whether properties have been set properly
    assert mesher.minimum_angle == 10.0
    assert mesher.conforming_delaunay is False
    assert mesher.suppress_exact_arithmetic is True
    assert mesher.maximum_steiner_points == 10
    assert mesher.delaunay_algorithm == pm.DelaunayAlgorithm.SWEEPLINE
    assert mesher.consistency_check is True

    # Check whether the repr method works
    assert isinstance(mesher.__repr__(), str)

    with pytest.raises(TypeError):
        mesher.conforming_delaunay = "a"

    with pytest.raises(TypeError):
        mesher.suppress_exact_arithmetic = "a"

    with pytest.raises(TypeError):
        mesher.maximum_steiner_points = "a"

    with pytest.raises(ValueError):
        mesher.delaunay_algorithm = "a"

    with pytest.raises(TypeError):
        mesher.consistency_check = "a"


def test_gmsh_properties():
    gdf = gpd.GeoDataFrame(geometry=[donut])
    gdf["cellsize"] = 1.0
    mesher = pm.GmshMesher._force_init(gdf)

    # Set default values for meshing parameters
    mesher.mesh_algorithm = pm.MeshAlgorithm.FRONTAL_DELAUNAY
    mesher.recombine_all = False
    mesher.mesh_size_extend_from_boundary = False
    mesher.mesh_size_from_points = False
    mesher.mesh_size_from_curvature = True
    mesher.field_combination = pm.FieldCombination.MAX
    mesher.subdivision_algorithm = pm.SubdivisionAlgorithm.BARYCENTRIC
    mesher.general_verbosity = pm.GeneralVerbosity.ERRORS

    assert mesher.mesh_algorithm == pm.MeshAlgorithm.FRONTAL_DELAUNAY
    assert mesher.recombine_all is False
    assert mesher.mesh_size_extend_from_boundary is False
    assert mesher.mesh_size_from_points is False
    assert mesher.mesh_size_from_curvature is True
    assert mesher.field_combination == pm.FieldCombination.MAX
    assert mesher.subdivision_algorithm == pm.SubdivisionAlgorithm.BARYCENTRIC
    assert mesher.general_verbosity == pm.GeneralVerbosity.ERRORS

    # Check whether the repr method works
    assert isinstance(mesher.__repr__(), str)

    with pytest.raises(ValueError):
        mesher.mesh_algorithm = "a"

    with pytest.raises(TypeError):
        mesher.recombine_all = "a"

    with pytest.raises(TypeError):
        mesher.mesh_size_extend_from_boundary = "a"

    with pytest.raises(TypeError):
        mesher.mesh_size_from_points = "a"

    with pytest.raises(TypeError):
        mesher.mesh_size_from_curvature = "a"

    with pytest.raises(ValueError):
        mesher.field_combination = "a"

    with pytest.raises(ValueError):
        mesher.subdivision_algorithm = "a"

    with pytest.raises(ValueError):
        mesher.general_verbosity = "a"


def test_gmsh_write(tmp_path):
    gdf = gpd.GeoDataFrame(geometry=[donut])
    gdf["cellsize"] = 1.0
    mesher = pm.GmshMesher._force_init(gdf)
    path = tmp_path / "a.msh"
    mesher.write(path)
    assert path.exists()


@pytest.mark.parametrize("read_config_files", [True, False])
@pytest.mark.parametrize("interruptible", [True, False])
def test_gmsh_initialization_kwargs(read_config_files, interruptible):
    gdf = gpd.GeoDataFrame(geometry=[donut])
    gdf["cellsize"] = 1.0
    mesher = pm.GmshMesher._force_init(
        gdf, read_config_files=read_config_files, interruptible=interruptible
    )
    vertices, triangles = mesher.generate()
    mesh_area = area(vertices, triangles).sum()
    assert np.allclose(mesh_area, donut.area)


def test_generate_geodataframe():
    gdf = gpd.GeoDataFrame(geometry=[donut])
    gdf["cellsize"] = 1.0

    result = pm.TriangleMesher(gdf).generate_geodataframe()
    assert isinstance(result, gpd.GeoDataFrame)
    assert np.allclose(result.area.sum(), donut.area)

    mesher = pm.GmshMesher._force_init(gdf)
    result = mesher.generate_geodataframe()
    assert isinstance(result, gpd.GeoDataFrame)
    assert np.allclose(result.area.sum(), donut.area)

    # Make sure the geodataframe logic can deal with quads as well.
    mesher.recombine_all = True
    result = mesher.generate_geodataframe()
    assert isinstance(result, gpd.GeoDataFrame)
    assert np.allclose(result.area.sum(), donut.area)



import re
import textwrap

import pytest

from pandamesh.enum_base import FlexibleEnum, _show_options


class Color(FlexibleEnum):
    RED = 1
    GREEN = 2
    BLUE = 3


def test_show_options():
    _show_options(Color) == textwrap.dedent(
        """
        * RED
        * GREEN
        * BLUE
    """
    )


def test_check_options():
    assert Color(1) == Color.RED
    assert Color(Color.RED) == Color.RED
    assert Color.from_value(Color.RED) == Color.RED
    assert Color.from_value("RED") == Color.RED
    expected = (
        "'YELLOW' is not a valid Color. Valid options are:\n * RED\n * GREEN\n * BLUE"
    )
    with pytest.raises(ValueError, match=re.escape(expected)):
        Color.from_value("YELLOW")

    expected = expected.replace("'YELLOW'", "YELLOW")
    with pytest.raises(AttributeError, match=re.escape(expected)):
        Color.YELLOW

    expected = expected.replace("YELLOW", "0")
    with pytest.raises(ValueError, match=re.escape(expected)):
        Color.from_value(0)



from typing import NamedTuple

import numpy as np
import shapely

from pandamesh.common import FloatArray, GeometryArray, IntArray

try:
    from numba import njit  # pragma: no cover
except ImportError:

    def njit(*args, **kwargs):  # pragma: no cover
        # Dummy decorator when numba is not available.
        def decorator(func):
            return func

        return decorator


class MatrixCSR(NamedTuple):
    """
    Compressed Sparse Row matrix. The row indices are compressed; all values
    must therefore be sorted by row number. More or less matches the
    scipy.sparse.csr_matrix.

    NamedTuple for easy ingestion by numba.

    Parameters
    ----------
    data: np.ndarray of floats
        The values of the matrix.
    indices: np.ndarray of integers
        The column numbers of the CSR format.
    indptr: inp.ndarray of integers
        The row index CSR pointer array.
        Values for row i (target index i) are stored in:
        indices[indptr[i]: indptr[i + 1]]
    n: int
        The number of rows.
    m: int
        The number of columns.
    nnz: int
        The number of non-zero values.
    """

    data: FloatArray
    indices: IntArray
    indptr: IntArray
    n: int
    m: int
    nnz: int

    @staticmethod
    def from_triplet(
        row: IntArray, col: IntArray, data: np.ndarray, n: int, m: int
    ) -> "MatrixCSR":
        i = np.cumsum(np.bincount(row, minlength=n))
        indptr = np.empty(i.size + 1, dtype=int)
        indptr[0] = 0
        indptr[1:] = i
        return MatrixCSR(
            data=data,
            indices=col,
            indptr=indptr,
            n=n,
            m=m,
            nnz=data.size,
        )


@njit(inline="always")
def row_slice(A, row: int) -> slice:
    """Return the indices or data slice of a single row."""
    start = A.indptr[row]
    end = A.indptr[row + 1]
    return slice(start, end)


@njit(inline="always")
def columns_and_values(A, slice):
    return zip(A.indices[slice], A.data[slice])


@njit(cache=True)
def _snap_to_nearest(A: MatrixCSR, snap_candidates: IntArray, max_distance) -> IntArray:
    """
    Find a closest target for each node.

    The kD tree distance matrix will have stored for each node the other nodes
    that are within snapping distance. These are the rows in the sparse matrix
    that have more than one entry: the snap_candidates.

    The first condition for a point to become a TARGET is if it hasn't been
    connected to another point yet, i.e. it is UNVISITED. Once a point becomes
    an TARGET, it looks for nearby points within the max_distance. These nearby
    points are connected if: they are UNVISITED (i.e. they don't have a target
    yet), or the current target is closer than the previous.
    """
    UNVISITED = -1
    TARGET = -2
    nearest = np.full(A.n, max_distance + 1.0)
    visited = np.full(A.n, UNVISITED)

    for i in snap_candidates:
        if visited[i] != UNVISITED:
            continue
        visited[i] = TARGET

        # Now iterate through every node j that is within max_distance of node i.
        for j, dist in columns_and_values(A, row_slice(A, i)):
            if i == j or visited[j] == TARGET:
                # Continue if we're looking at the distance to ourselves
                # (i==j), or other node is a target.
                continue
            if visited[j] == UNVISITED or dist < nearest[j]:
                # If unvisited node, or already visited but we're closer, set
                # to i.
                visited[j] = i
                nearest[j] = dist

    return visited


def distance_matrix(geometry: GeometryArray, max_distance: float) -> MatrixCSR:
    n = len(geometry)
    tree = shapely.STRtree(geometry)
    i, j = tree.query(geometry, distance=max_distance, predicate="dwithin")
    coords = shapely.get_coordinates(geometry)
    return MatrixCSR.from_triplet(
        row=i,
        col=j,
        data=np.linalg.norm(coords[i] - coords[j], axis=1),
        n=n,
        m=n,
    )


def snap_nodes(points: GeometryArray, max_distance: float) -> IntArray:
    """
    Snap neigbhoring vertices together that are located within a maximum
    snapping distance from each other.

    If vertices are located within a maximum distance, some of them are snapped
    to their neighbors ("targets"), thereby guaranteeing a minimum distance
    between nodes in the result. The determination of whether a point becomes a
    target itself or gets snapped to another point is primarily based on the
    order in which points are processed and their spatial relationships.

    This function also return an inverse index array. In case of a connectivity
    array, ``inverse`` can be used to index into, yielding the updated
    numbers. E.g.:

    ``updated_face_nodes = inverse[face_nodes]``

    Parameters
    ----------
    xy: nd array of floats of size (N, 2)
    max_distance: float

    Returns
    -------
    index: 1D array of ints of size M
        Which nodes to preserve.
    """
    # First, find all the points that lie within max_distance of each other
    A = distance_matrix(points, max_distance)
    should_snap = np.diff(A.indptr) > 1  # equal to: .getnnz(axis=1) > 1
    if should_snap.any():
        index = np.arange(A.n)
        visited = _snap_to_nearest(
            A=A,
            snap_candidates=index[should_snap],
            max_distance=max_distance,
        )
        targets = visited < 0  # i.e. still UNVISITED or TARGET valued.
        visited[targets] = index[targets]
        deduplicated = np.unique(visited)
        return deduplicated
    else:
        return np.arange(A.n)



from typing import Tuple, Union

import geopandas as gpd
import triangle

from pandamesh.common import (
    FloatArray,
    IntArray,
    central_origin,
    check_geodataframe,
    repr,
    separate_geodataframe,
    to_geodataframe,
    to_ugrid,
)
from pandamesh.triangle_enums import DelaunayAlgorithm
from pandamesh.triangle_geometry import collect_geometry, polygon_holes


class TriangleMesher:
    """
    Wrapper for the python bindings to Triangle. This class must be initialized
    with a geopandas GeoDataFrame containing at least one polygon, and a column
    named ``"cellsize"``.

    Optionally, multiple polygons with different cell sizes can be included in
    the geodataframe. These can be used to achieve local mesh refinement.

    Linestrings and points may also be included. The segments of linestrings
    will be directly forced into the triangulation. Points can also be forced
    into the triangulation. The cell size values associated with these
    geometries willl not be used.

    Triangle cannot automatically resolve overlapping polygons, or points
    located exactly on segments. During initialization, the geometries of
    the geodataframe are checked:

        * Polygons should not have any overlap with each other.
        * Linestrings should not intersect each other, unless the intersection
          vertex is present in both.
        * Every linestring should be fully contained by a single polygon;
          a linestring may not intersect two or more polygons.
        * Linestrings and points should not "touch" / be located on
          polygon borders.
        * Holes in polygons are fully supported, but they must not contain
          any linestrings or points.

    If such cases are detected, the initialization will error: use the
    :class:`pandamesh.Preprocessor` to clean up geometries beforehand.

    For more details on Triangle, see:
    https://www.cs.cmu.edu/~quake/triangle.defs.html

    Parameters
    ----------
    gdf: gpd.GeoDataFrame
        GeoDataFrame containing the vector geometry. Must contain a "cellsize"
        column.
    shift_origin: bool, optional, default is True.
        If True, temporarily shifts the coordinate system origin to the centroid
        of the geometry's bounding box during mesh generation. This helps mitigate
        floating-point precision issues. The resulting mesh vertices are
        automatically translated back to the original coordinate system.
    intersecting_edges: str, optional, default is "error"
        String indicating how to report unresolved line segment intersections:

        * "ignore": skip check.
        * "warning": emit a warning.
        * "error": raise a ValueError.

    """

    def __init__(
        self,
        gdf: gpd.GeoDataFrame,
        shift_origin: bool = True,
        intersecting_edges="error",
    ) -> None:
        check_geodataframe(gdf, {"geometry", "cellsize"}, check_index=True)
        gdf, self._xoff, self._yoff = central_origin(gdf, shift_origin)
        polygons, linestrings, points = separate_geodataframe(gdf, intersecting_edges)
        self.vertices, self.segments, self.regions = collect_geometry(
            polygons, linestrings, points
        )
        self.holes = polygon_holes(polygons)

        # Set default values for meshing parameters
        self.minimum_angle = 20.0
        self.conforming_delaunay = True
        self.suppress_exact_arithmetic = False
        self.maximum_steiner_points = None
        self.delaunay_algorithm = DelaunayAlgorithm.DIVIDE_AND_CONQUER
        self.consistency_check = False

    def __repr__(self):
        return repr(self)

    @property
    def minimum_angle(self) -> float:
        """
        Minimum allowed angle for any triangle in the mesh.

        See:
        https://www.cs.cmu.edu/~quake/triangle.q.html
        """
        return self._minimum_angle

    @minimum_angle.setter
    def minimum_angle(self, value: float):
        if not isinstance(value, float):
            raise TypeError("minimum angle must be a float")
        if value >= 34.0 or value <= 0.0:
            raise ValueError("minimum_angle should fall in the interval: (0.0, 34.0)")
        self._minimum_angle = value

    @property
    def conforming_delaunay(self) -> bool:
        """
        Conforming Delaunay: use this switch if you want all triangles in the
        mesh to be Delaunay, and not just constrained Delaunay; or if you want
        to ensure that all Voronoi vertices lie within the triangulation.
        """
        return self._conforming_delaunay

    @conforming_delaunay.setter
    def conforming_delaunay(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("conforming_delaunay must be a bool")
        self._conforming_delaunay = value

    @property
    def suppress_exact_arithmetic(self) -> bool:
        """
        Suppresses exact arithmetic.

        See:
        https://www.cs.cmu.edu/~quake/triangle.exact.html
        """
        return self._suppress_exact_arithmetic

    @suppress_exact_arithmetic.setter
    def suppress_exact_arithmetic(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("suppress_exact_arithmetic must be a bool")
        self._suppress_exact_arithmetic = value

    @property
    def maximum_steiner_points(self) -> int:
        """
        Specifies the maximum number of added Steiner points

        See:
        https://www.cs.cmu.edu/~quake/triangle.S.html
        """
        return self._maximum_steiner_points

    @maximum_steiner_points.setter
    def maximum_steiner_points(self, value: Union[int, None]):
        if not isinstance(value, (int, type(None))):
            raise TypeError("maximum_steiner_points must be an int or None")
        self._maximum_steiner_points = value

    @property
    def delaunay_algorithm(self) -> DelaunayAlgorithm:
        """
        Sets the Delaunay algorithm. Can be set to one of:
        :py:class:`pandamesh.DelaunayAlgorithm`:

        .. code::

            DIVIDE_AND_CONQUER = ""
            INCREMENTAL = "i"
            SWEEPLINE = "F"

        """
        return self._delaunay_algorithm

    @delaunay_algorithm.setter
    def delaunay_algorithm(self, value: Union[DelaunayAlgorithm, str]):
        value = DelaunayAlgorithm.from_value(value)
        self._delaunay_algorithm = value

    @property
    def consistency_check(self) -> bool:
        """
        Check the consistency of the final mesh. Uses exact arithmetic for
        checking, even if ``suppress_exact_arithmetic`` is set to ``False``.
        Useful if you suspect Triangle is buggy.
        """
        return self._consistency_check

    @consistency_check.setter
    def consistency_check(self, value: bool):
        if not isinstance(value, int):
            raise TypeError("consistency_check must be a bool")
        self._consistency_check = value

    def generate(self) -> Tuple[FloatArray, IntArray]:
        """
        Generate a mesh of triangles.

        Returns
        -------
        vertices: np.ndarray of floats with shape ``(n_vertex, 2)``
        triangles: np.ndarray of integers with shape ``(n_triangle, 3)``
        """
        options = (
            "p"
            f"q{self._minimum_angle}"
            "a"
            f"{'D' if self._conforming_delaunay else ''}"
            f"{'X' if self._suppress_exact_arithmetic else ''}"
            f"{'S' + str(self._maximum_steiner_points) if self._maximum_steiner_points is not None else ''}"
            f"{self._delaunay_algorithm.value}"
            f"{'C' if self.consistency_check else ''}"
        )

        tri = {"vertices": self.vertices, "segments": self.segments}
        if self.holes is not None:
            tri["holes"] = self.holes
        if len(self.regions) > 0:
            tri["regions"] = self.regions

        result = triangle.triangulate(tri=tri, opts=options)
        vertices = result["vertices"]
        vertices[:, 0] += self._xoff
        vertices[:, 1] += self._yoff
        return vertices, result["triangles"]

    def generate_geodataframe(self) -> gpd.GeoDataFrame:
        """
        Generate a mesh and return it as a geopandas GeoDataFrame.

        Returns
        -------
        mesh: geopandas.GeoDataFrame
        """
        return to_geodataframe(*self.generate())

    def generate_ugrid(self) -> "xugrid.Ugrid2d":  # type: ignore # noqa  pragma: no cover
        """
        Generate a mesh and return it as an xugrid Ugrid2d.

        Returns
        -------
        mesh: xugrid.Ugrid2d
        """
        return to_ugrid(*self.generate())



from pandamesh import data
from pandamesh.common import find_edge_intersections
from pandamesh.gmsh_enums import (
    FieldCombination,
    GeneralVerbosity,
    MeshAlgorithm,
    SubdivisionAlgorithm,
)
from pandamesh.gmsh_mesher import GmshMesher, gmsh_env
from pandamesh.plot import plot
from pandamesh.preprocessor import Preprocessor
from pandamesh.triangle_enums import DelaunayAlgorithm
from pandamesh.triangle_mesher import TriangleMesher

__version__ = "0.2.0"


__all__ = (
    "data",
    "FieldCombination",
    "GeneralVerbosity",
    "GmshMesher",
    "MeshAlgorithm",
    "SubdivisionAlgorithm",
    "gmsh_env",
    "plot",
    "DelaunayAlgorithm",
    "TriangleMesher",
    "Preprocessor",
    "find_edge_intersections",
)
